import warnings
from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Annotated, Any, NamedTuple, Optional, TypeAlias

from dagster_shared.record import ImportFrom
from dagster_shared.utils.cached_method import get_cached_method_cache
from typing_extensions import Self

import dagster._check as check
from dagster._annotations import deprecated, preview, public
from dagster._core.definitions import AssetSelection
from dagster._core.definitions.asset_checks.asset_checks_definition import AssetChecksDefinition
from dagster._core.definitions.asset_key import AssetCheckKey
from dagster._core.definitions.asset_selection import CoercibleToAssetSelection
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec, map_asset_specs
from dagster._core.definitions.assets.definition.assets_definition import (
    AssetsDefinition,
    SourceAsset,
)
from dagster._core.definitions.assets.definition.cacheable_assets_definition import (
    CacheableAssetsDefinition,
)
from dagster._core.definitions.assets.graph.asset_graph import AssetGraph
from dagster._core.definitions.decorators import repository
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.executor_definition import ExecutorDefinition
from dagster._core.definitions.job_definition import JobDefinition, default_job_io_manager
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.metadata import RawMetadataMapping, normalize_metadata
from dagster._core.definitions.metadata.metadata_value import (
    CodeLocationReconstructionMetadataValue,
    MetadataValue,
)
from dagster._core.definitions.partitions.partitioned_schedule import (
    UnresolvedPartitionedAssetScheduleDefinition,
)
from dagster._core.definitions.repository_definition import (
    SINGLETON_REPOSITORY_NAME,
    RepositoryDefinition,
)
from dagster._core.definitions.schedule_definition import ScheduleDefinition
from dagster._core.definitions.sensor_definition import SensorDefinition
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.definitions.utils import dedupe_object_refs
from dagster._core.errors import DagsterInvariantViolationError
from dagster._core.execution.build_resources import wrap_resources_for_execution
from dagster._core.execution.with_resources import with_resources
from dagster._core.executor.base import Executor
from dagster._core.instance import DagsterInstance
from dagster._record import IHaveNew, copy, record_custom, replace
from dagster._utils.cached_method import cached_method
from dagster._utils.warnings import disable_dagster_warnings

if TYPE_CHECKING:
    from dagster._core.storage.asset_value_loader import AssetValueLoader
    from dagster.components.core.component_tree import ComponentTree


TAssets: TypeAlias = (
    Iterable[AssetsDefinition | AssetSpec | SourceAsset | CacheableAssetsDefinition] | None
)
TSchedules: TypeAlias = (
    Iterable[ScheduleDefinition | UnresolvedPartitionedAssetScheduleDefinition] | None
)
TSensors: TypeAlias = Iterable[SensorDefinition] | None
TJob: TypeAlias = JobDefinition | UnresolvedAssetJobDefinition
TJobs: TypeAlias = Iterable[TJob] | None
TAssetChecks: TypeAlias = Iterable[AssetsDefinition] | None


@public
def create_repository_using_definitions_args(
    name: str,
    assets: TAssets = None,
    schedules: TSchedules = None,
    sensors: TSensors = None,
    jobs: TJobs = None,
    resources: Mapping[str, Any] | None = None,
    executor: ExecutorDefinition | Executor | None = None,
    loggers: Mapping[str, LoggerDefinition] | None = None,
    asset_checks: TAssetChecks = None,
) -> RepositoryDefinition:
    """Create a named repository using the same arguments as :py:class:`Definitions`. In older
    versions of Dagster, repositories were the mechanism for organizing assets, schedules, sensors,
    and jobs. There could be many repositories per code location. This was a complicated ontology but
    gave users a way to organize code locations that contained large numbers of heterogenous definitions.

    As a stopgap for those who both want to 1) use the new :py:class:`Definitions` API and 2) but still
    want multiple logical groups of assets in the same code location, we have introduced this function.

    Example usage:

    .. code-block:: python

        named_repo = create_repository_using_definitions_args(
            name="a_repo",
            assets=[asset_one, asset_two],
            schedules=[a_schedule],
            sensors=[a_sensor],
            jobs=[a_job],
            resources={
                "a_resource": some_resource,
            }
        )

    """
    return _create_repository_using_definitions_args(
        name=name,
        assets=assets,
        schedules=schedules,
        sensors=sensors,
        jobs=jobs,
        resources=resources,
        executor=executor,
        loggers=loggers,
        asset_checks=asset_checks,
    )


class _AttachedObjects(NamedTuple):
    jobs: Iterable[JobDefinition | UnresolvedAssetJobDefinition]
    schedules: Iterable[ScheduleDefinition | UnresolvedPartitionedAssetScheduleDefinition]
    sensors: Iterable[SensorDefinition]


def _io_manager_needs_replacement(job: JobDefinition, resource_defs: Mapping[str, Any]) -> bool:
    """Explicitly replace the default IO manager in jobs that don't specify one, if a top-level
    I/O manager is provided to Definitions.
    """
    return (
        job.resource_defs.get("io_manager") == default_job_io_manager
        and "io_manager" in resource_defs
    )


def _attach_resources_to_jobs_and_instigator_jobs(
    jobs: Iterable[JobDefinition | UnresolvedAssetJobDefinition] | None,
    schedules: Iterable[ScheduleDefinition | UnresolvedPartitionedAssetScheduleDefinition] | None,
    sensors: Iterable[SensorDefinition] | None,
    resource_defs: Mapping[str, Any],
) -> _AttachedObjects:
    """Given a list of jobs, schedules, and sensors along with top-level resource definitions,
    attach the resource definitions to the jobs, schedules, and sensors which require them.
    """
    jobs = jobs or []
    schedules = schedules or []
    sensors = sensors or []

    # Add jobs in schedules and sensors as well
    jobs = [
        *jobs,
        *[
            schedule.job
            for schedule in schedules
            if isinstance(schedule, ScheduleDefinition)
            and schedule.target.has_job_def
            and isinstance(schedule.job, (JobDefinition, UnresolvedAssetJobDefinition))
        ],
        *[
            target.job_def
            for sensor in sensors
            for target in sensor.targets
            if target.has_job_def
            and isinstance(target.job_def, (JobDefinition, UnresolvedAssetJobDefinition))
        ],
    ]
    # Dedupe
    jobs = list({id(job): job for job in jobs}.values())

    # Find unsatisfied jobs
    unsatisfied_jobs = [
        job
        for job in jobs
        if isinstance(job, JobDefinition)
        and (
            job.is_missing_required_resources() or _io_manager_needs_replacement(job, resource_defs)
        )
    ]

    # Create a mapping of job id to a version of the job with the resource defs bound
    unsatisfied_job_to_resource_bound_job = {
        id(job): job.with_top_level_resources(
            {
                **resource_defs,
                **job.resource_defs,
                # special case for IO manager - the job-level IO manager does not take precedence
                # if it is the default and a top-level IO manager is provided
                **(
                    {"io_manager": resource_defs["io_manager"]}
                    if _io_manager_needs_replacement(job, resource_defs)
                    else {}
                ),
            }
        )
        for job in jobs
        if job in unsatisfied_jobs
    }

    # Update all jobs to use the resource bound version
    jobs_with_resources = [
        unsatisfied_job_to_resource_bound_job[id(job)] if job in unsatisfied_jobs else job
        for job in jobs
    ]

    # Update all schedules and sensors to use the resource bound version
    updated_schedules = [
        (
            schedule.with_updated_job(unsatisfied_job_to_resource_bound_job[id(schedule.job)])
            if (
                isinstance(schedule, ScheduleDefinition)
                and schedule.target.has_job_def
                and schedule.job in unsatisfied_jobs
            )
            else schedule
        )
        for schedule in schedules
    ]
    updated_sensors = [
        (
            sensor.with_updated_jobs(
                [
                    (
                        unsatisfied_job_to_resource_bound_job[id(job)]
                        if job in unsatisfied_jobs
                        else job
                    )
                    for job in sensor.jobs
                ]
            )
            if any(target.has_job_def for target in sensor.targets)
            and any(job in unsatisfied_jobs for job in sensor.jobs)
            else sensor
        )
        for sensor in sensors
    ]

    return _AttachedObjects(jobs_with_resources, updated_schedules, updated_sensors)


def _create_repository_using_definitions_args(
    name: str,
    assets: TAssets = None,
    schedules: TSchedules = None,
    sensors: TSensors = None,
    jobs: TJobs = None,
    resources: Mapping[str, Any] | None = None,
    executor: ExecutorDefinition | Executor | None = None,
    loggers: Mapping[str, LoggerDefinition] | None = None,
    asset_checks: TAssetChecks = None,
    metadata: RawMetadataMapping | None = None,
    component_tree: Optional["ComponentTree"] = None,
) -> RepositoryDefinition:
    # First, dedupe all definition types.
    sensors = dedupe_object_refs(sensors)
    jobs = dedupe_object_refs(jobs)
    assets = _canonicalize_specs_to_assets_defs(dedupe_object_refs(assets))
    schedules = dedupe_object_refs(schedules)
    asset_checks = dedupe_object_refs(asset_checks)

    executor_def = (
        executor
        if isinstance(executor, ExecutorDefinition) or executor is None
        else ExecutorDefinition.hardcoded_executor(executor)
    )

    resource_defs = wrap_resources_for_execution(resources)

    # Binds top-level resources to jobs and any jobs attached to schedules or sensors
    (
        jobs_with_resources,
        schedules_with_resources,
        sensors_with_resources,
    ) = _attach_resources_to_jobs_and_instigator_jobs(jobs, schedules, sensors, resource_defs)

    @repository(
        name=name,
        default_executor_def=executor_def,
        default_logger_defs=loggers,
        metadata=metadata,
        _top_level_resources=resource_defs,
        _component_tree=component_tree,
    )
    def created_repo():
        return [
            *with_resources(assets, resource_defs),
            *with_resources(asset_checks or [], resource_defs),
            *(schedules_with_resources),
            *(sensors_with_resources),
            *(jobs_with_resources),
        ]

    return created_repo


def _canonicalize_specs_to_assets_defs(
    assets: Iterable[AssetsDefinition | AssetSpec | SourceAsset | CacheableAssetsDefinition],
) -> Iterable[AssetsDefinition | SourceAsset | CacheableAssetsDefinition]:
    asset_specs_by_partitions_def = defaultdict(list)
    for obj in assets:
        if isinstance(obj, AssetSpec):
            asset_specs_by_partitions_def[obj.partitions_def].append(obj)
    result = [obj for obj in assets if not isinstance(obj, AssetSpec)]

    for specs in asset_specs_by_partitions_def.values():
        with disable_dagster_warnings():
            result.append(AssetsDefinition(specs=specs))

    return result


@deprecated(
    breaking_version="2.0",
    additional_warn_text=(
        "Instantiations can be removed. Since it's behavior is now the default, this class is now a"
        " no-op."
    ),
)
class BindResourcesToJobs(list):
    """Used to instruct Dagster to bind top-level resources to jobs and any jobs attached to schedules
    and sensors. Now deprecated since this behavior is the default.
    """


@record_custom
@public
class Definitions(IHaveNew):
    """A set of definitions explicitly available and loadable by Dagster tools.

    Parameters:
        assets (Optional[Iterable[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]]):
            A list of assets. Assets can be created by annotating
            a function with :py:func:`@asset <asset>` or
            :py:func:`@observable_source_asset <observable_source_asset>`.
            Or they can by directly instantiating :py:class:`AssetsDefinition`,
            :py:class:`SourceAsset`, or :py:class:`CacheableAssetsDefinition`.

        asset_checks (Optional[Iterable[AssetChecksDefinition]]):
            A list of asset checks.

        schedules (Optional[Iterable[Union[ScheduleDefinition, UnresolvedPartitionedAssetScheduleDefinition]]]):
            List of schedules.

        sensors (Optional[Iterable[SensorDefinition]]):
            List of sensors, typically created with :py:func:`@sensor <sensor>`.

        jobs (Optional[Iterable[Union[JobDefinition, UnresolvedAssetJobDefinition]]]):
            List of jobs. Typically created with :py:func:`define_asset_job <define_asset_job>`
            or with :py:func:`@job <job>` for jobs defined in terms of ops directly.
            Jobs created with :py:func:`@job <job>` must already have resources bound
            at job creation time. They do not respect the `resources` argument here.

        resources (Optional[Mapping[str, Any]]): Dictionary of resources to bind to assets.
            The resources dictionary takes raw Python objects,
            not just instances of :py:class:`ResourceDefinition`. If that raw object inherits from
            :py:class:`IOManager`, it gets coerced to an :py:class:`IOManagerDefinition`.
            Any other object is coerced to a :py:class:`ResourceDefinition`.
            These resources will be automatically bound
            to any assets passed to this Definitions instance using
            :py:func:`with_resources <with_resources>`. Assets passed to Definitions with
            resources already bound using :py:func:`with_resources <with_resources>` will
            override this dictionary.

        executor (Optional[Union[ExecutorDefinition, Executor]]):
            Default executor for jobs. Individual jobs can override this and define their own executors
            by setting the executor on :py:func:`@job <job>` or :py:func:`define_asset_job <define_asset_job>`
            explicitly. This executor will also be used for materializing assets directly
            outside of the context of jobs. If an :py:class:`Executor` is passed, it is coerced into
            an :py:class:`ExecutorDefinition`.

        loggers (Optional[Mapping[str, LoggerDefinition]):
            Default loggers for jobs. Individual jobs
            can define their own loggers by setting them explictly.

        metadata (Optional[MetadataMapping]):
            Arbitrary metadata for the Definitions. Not displayed in the UI but accessible on
            the Definitions instance at runtime.

        component_tree (Optional[ComponentTree]):
            Information about the Components that were used to construct part of this
            Definitions object.


    Example usage:

    .. code-block:: python

        Definitions(
            assets=[asset_one, asset_two],
            schedules=[a_schedule],
            sensors=[a_sensor],
            jobs=[a_job],
            resources={
                "a_resource": some_resource,
            },
            asset_checks=[asset_one_check_one]
        )

    Dagster separates user-defined code from system tools such the web server and
    the daemon. Rather than loading code directly into process, a tool such as the
    webserver interacts with user-defined code over a serialization boundary.

    These tools must be able to locate and load this code when they start. Via CLI
    arguments or config, they specify a Python module to inspect.

    A Python module is loadable by Dagster tools if there is a top-level variable
    that is an instance of :py:class:`Definitions`.
    """

    assets: TAssets = None
    schedules: TSchedules = None
    sensors: TSensors = None
    jobs: TJobs = None
    resources: Mapping[str, Any] | None = None
    executor: ExecutorDefinition | Executor | None = None
    loggers: Mapping[str, LoggerDefinition] | None = None
    # There's a bug that means that sometimes it's Dagster's fault when AssetsDefinitions are
    # passed here instead of AssetChecksDefinitions: https://github.com/dagster-io/dagster/issues/22064.
    # After we fix the bug, we should remove AssetsDefinition from the set of accepted types.
    asset_checks: TAssetChecks = None
    metadata: Mapping[str, MetadataValue]
    component_tree: (
        Annotated["ComponentTree", ImportFrom("dagster.components.core.component_tree")] | None
    )

    def __new__(
        cls,
        assets: TAssets = None,
        schedules: TSchedules = None,
        sensors: TSensors = None,
        jobs: TJobs = None,
        resources: Mapping[str, Any] | None = None,
        executor: ExecutorDefinition | Executor | None = None,
        loggers: Mapping[str, LoggerDefinition] | None = None,
        asset_checks: TAssetChecks = None,
        metadata: RawMetadataMapping | None = None,
        component_tree: Optional["ComponentTree"] = None,
    ):
        instance = super().__new__(
            cls,
            assets=assets,
            schedules=schedules,
            sensors=sensors,
            jobs=jobs,
            resources=resources,
            executor=executor,
            loggers=loggers,
            asset_checks=asset_checks,
            metadata=normalize_metadata(check.opt_mapping_param(metadata, "metadata")),
            component_tree=component_tree,
        )

        check.invariant(
            not instance.has_resolved_repository_def(),
            "Definitions object should not have been resolved",
        )

        return instance

    @public
    def get_job_def(self, name: str) -> JobDefinition:
        """Get a job definition by name. This will only return a `JobDefinition` if it was directly passed in to the `Definitions` object.

        If that is not found, the Definitions object is resolved (transforming UnresolvedAssetJobDefinitions to JobDefinitions and an example). It
        also finds jobs passed to sensors and schedules and retrieves them from the repository.

        After dagster 1.11, this resolution step will not happen, and will throw an error if the job is not found.
        """
        found_direct = False
        for job in self.jobs or []:
            if job.name == name:
                if isinstance(job, JobDefinition):
                    found_direct = True

        if not found_direct:
            warning = self.dig_for_warning(name)
            if warning:
                warnings.warn(warning)
            else:
                warnings.warn(
                    f"JobDefinition with name {name} directly passed to Definitions not found, "
                    "will attempt to resolve to a JobDefinition. "
                    "This will be an error in a future release and will require a call to "
                    "resolve_job_def in dagster 1.11. "
                )

        return self.resolve_job_def(name)

    def resolve_job_def(self, name: str) -> JobDefinition:
        """Resolve a job definition by name. If you passed in an :py:class:`UnresolvedAssetJobDefinition`
        (return value of :py:func:`define_asset_job`) it will be resolved to a :py:class:`JobDefinition` when returned
        from this function, with all resource dependencies fully resolved.
        """
        check.str_param(name, "name")
        return self.get_repository_def().get_job(name)

    def dig_for_warning(self, name: str) -> str | None:
        for job in self.jobs or []:
            if job.name == name:
                if isinstance(job, JobDefinition):
                    return None
                return (
                    f"Found asset job named {job.name} of type {type(job)} passed to `jobs` parameter. Starting in "
                    "dagster 1.11, you must now use Definitions.resolve_job_def to correctly "
                    "retrieve this job definition."
                )

        for sensor in self.sensors or []:
            for job in sensor.jobs:
                if job.name == name:
                    return (
                        f"Found job or graph named {job.name} passed to sensor named {sensor.name} "
                        "that was passed to Definitions in the sensors param. Starting in dagster 1.11, "
                        "you must call Definitions.resolve_job_def to retrieve this job definition."
                    )

        for schedule in self.schedules or []:
            job = schedule.job
            if job.name == name:
                return (
                    f"Found job named {job.name} passed to schedule named {schedule.name} "
                    "that was passed to Definitions in the schedules param. Starting in dagster 1.11, "
                    "you must call Definitions.resolve_job_def to retrieve this job definition."
                )
        return None

    @public
    def get_sensor_def(self, name: str) -> SensorDefinition:
        """Get a :py:class:`SensorDefinition` by name.
        If your passed-in sensor had resource dependencies, or the job targeted by the sensor had
        resource dependencies, those resource dependencies will be fully resolved on the returned object.
        """
        warnings.warn(
            "Starting in dagster 1.11, get_sensor_def will return a SensorDefinition without resolving resource dependencies on it or its target."
        )

        return self.resolve_sensor_def(name)

    # TODO: after dagster 1.11, this will become the implementation of get_sensor_def -- schrockn 2025-06-02
    def get_unresolved_sensor_def(self, name: str) -> SensorDefinition:
        for sensor in self.sensors or []:
            if sensor.name == name:
                return sensor
        raise ValueError(f"SensorDefinition with name {name} not found")

    def resolve_sensor_def(self, name: str) -> SensorDefinition:
        check.str_param(name, "name")
        return self.get_repository_def().get_sensor_def(name)

    @public
    def get_schedule_def(self, name: str) -> ScheduleDefinition:
        """Get a :py:class:`ScheduleDefinition` by name.
        If your passed-in schedule had resource dependencies, or the job targeted by the schedule had
        resource dependencies, those resource dependencies will be fully resolved on the returned object.
        """
        warnings.warn(
            "Starting in dagster 1.11, get_schedule_def will return a ScheduleDefinition without resolving resource dependencies on it or its target."
        )
        return self.resolve_schedule_def(name)

    # TODO: after dagster 1.11, this will become the implementation of get_schedule_def -- schrockn 2025-06-02
    def get_unresolved_schedule_def(self, name: str) -> ScheduleDefinition:
        for schedule in self.schedules or []:
            if schedule.name == name:
                if isinstance(schedule, ScheduleDefinition):
                    return schedule
                raise ValueError(
                    f"ScheduleDefinition with name {name} is an UnresolvedPartitionedAssetScheduleDefinition, which is not supported in get_unresolved_schedule_def"
                )

        raise ValueError(f"ScheduleDefinition with name {name} not found")

    def resolve_schedule_def(self, name: str) -> ScheduleDefinition:
        check.str_param(name, "name")
        return self.get_repository_def().get_schedule_def(name)

    @public
    def load_asset_value(
        self,
        asset_key: CoercibleToAssetKey,
        *,
        python_type: type | None = None,
        instance: DagsterInstance | None = None,
        partition_key: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> object:
        """Load the contents of an asset as a Python object.

        Invokes `load_input` on the :py:class:`IOManager` associated with the asset.

        If you want to load the values of multiple assets, it's more efficient to use
        :py:meth:`~dagster.Definitions.get_asset_value_loader`, which avoids spinning up
        resources separately for each asset.

        Args:
            asset_key (Union[AssetKey, Sequence[str], str]): The key of the asset to load.
            python_type (Optional[Type]): The python type to load the asset as. This is what will
                be returned inside `load_input` by `context.dagster_type.typing_type`.
            partition_key (Optional[str]): The partition of the asset to load.
            metadata (Optional[Dict[str, Any]]): Input metadata to pass to the :py:class:`IOManager`
                (is equivalent to setting the metadata argument in `In` or `AssetIn`).

        Returns:
            The contents of an asset as a Python object.
        """
        return self.get_repository_def().load_asset_value(
            asset_key=asset_key,
            python_type=python_type,
            instance=instance,
            partition_key=partition_key,
            metadata=metadata,
        )

    @public
    def get_asset_value_loader(self, instance: DagsterInstance | None = None) -> "AssetValueLoader":
        """Returns an object that can load the contents of assets as Python objects.

        Invokes `load_input` on the :py:class:`IOManager` associated with the assets. Avoids
        spinning up resources separately for each asset.

        Usage:

        .. code-block:: python

            with defs.get_asset_value_loader() as loader:
                asset1 = loader.load_asset_value("asset1")
                asset2 = loader.load_asset_value("asset2")
        """
        return self.get_repository_def().get_asset_value_loader(
            instance=instance,
        )

    def resolve_all_job_defs(self) -> Sequence[JobDefinition]:
        """Get all the Job definitions in the project.
        This includes both jobs passed into the Definitions object and any implicit jobs created.
        All jobs returned from this function will have all resource dependencies resolved.
        """
        return self.get_repository_def().get_all_jobs()

    def has_implicit_global_asset_job_def(self) -> bool:
        return self.get_repository_def().has_implicit_global_asset_job_def()

    def get_implicit_global_asset_job_def(self) -> JobDefinition:
        warnings.warn(
            "This will be renamed to resolve_implicit_global_asset_job_def in dagster 1.11"
        )
        return self.get_repository_def().get_implicit_global_asset_job_def()

    def resolve_implicit_global_asset_job_def(self) -> JobDefinition:
        """A useful conveninence method when there is a single defined global asset job.
        This occurs when all assets in the project use a single partitioning scheme.
        If there are multiple partitioning schemes you must use get_implicit_job_def_for_assets
        instead to access to the correct implicit asset one.
        """
        return self.get_repository_def().get_implicit_global_asset_job_def()

    def resolve_implicit_job_def_def_for_assets(
        self, asset_keys: Iterable[AssetKey]
    ) -> JobDefinition | None:
        return self.get_repository_def().get_implicit_job_def_for_assets(asset_keys)

    def get_assets_def(self, key: CoercibleToAssetKey) -> AssetsDefinition:
        key = AssetKey.from_coercible(key)
        # Sadly both self.assets and self.asset_checks can contain either AssetsDefinition or AssetChecksDefinition
        # objects. We need to check both collections and exclude the AssetChecksDefinition objects.
        for asset in [*(self.assets or []), *(self.asset_checks or [])]:
            if isinstance(asset, AssetsDefinition) and not isinstance(asset, AssetChecksDefinition):
                if key in asset.keys:
                    return asset

        warnings.warn(
            f"Could not find assets_def with key {key} directly passed to Definitions. This will be an error starting in 1.11 and will require a call to resolve_assets_def in dagster 1.11."
        )

        return self.resolve_assets_def(key)

    def get_asset_checks_def(self, key: AssetCheckKey) -> AssetChecksDefinition:
        for possible_assets_check_def in [*(self.assets or []), *(self.asset_checks or [])]:
            if (
                isinstance(possible_assets_check_def, AssetChecksDefinition)
                and key in possible_assets_check_def.asset_and_check_keys
            ):
                return possible_assets_check_def

        raise DagsterInvariantViolationError(f"Could not find asset checks defs for {key}")

    def resolve_assets_def(self, key: CoercibleToAssetKey) -> AssetsDefinition:
        asset_key = AssetKey.from_coercible(key)
        for assets_def in self.resolve_asset_graph().assets_defs:
            if asset_key in assets_def.keys:
                return assets_def

        raise DagsterInvariantViolationError(f"Could not find asset {asset_key}")

    @cached_method
    def get_repository_def(self) -> RepositoryDefinition:
        """Definitions is implemented by wrapping RepositoryDefinition. Get that underlying object
        in order to access any functionality which is not exposed on Definitions.
        """
        return _create_repository_using_definitions_args(
            name=SINGLETON_REPOSITORY_NAME,
            assets=self.assets,
            schedules=self.schedules,
            sensors=self.sensors,
            jobs=self.jobs,
            resources=self.resources,
            executor=self.executor,
            loggers=self.loggers,
            asset_checks=self.asset_checks,
            metadata=self.metadata,
            component_tree=self.component_tree,
        )

    def resolve_asset_graph(self) -> AssetGraph:
        """Get the AssetGraph for this set of definitions."""
        return self.get_repository_def().asset_graph

    @public
    @staticmethod
    def validate_loadable(defs: "Definitions") -> None:
        """Validates that the enclosed definitions will be loadable by Dagster:
        - No assets have conflicting keys.
        - No jobs, sensors, or schedules have conflicting names.
        - All asset jobs can be resolved.
        - All resource requirements are satisfied.
        - All partition mappings are valid.

        Meant to be used in unit tests.

        Raises an error if any of the above are not true.
        """
        defs.get_repository_def().validate_loadable()

    @staticmethod
    def merge_unbound_defs(*def_sets: "Definitions") -> "Definitions":
        """Merges multiple Definitions objects into a single Definitions object.

        Asserts that input Definitions objects have not yet had their asset graphs resolved,
        intended for internal use-cases to safeguard against unnecessarily resolving subgraphs.
        """
        for i, def_set in enumerate(def_sets):
            check.invariant(
                not def_set.has_resolved_repository_def(),
                f"Definitions object {i} has previously been resolved."
                " merge_unbound_defs should only be used on definitions that have not been resolved.",
            )

        return Definitions.merge(*def_sets)

    @public
    @staticmethod
    def merge(*def_sets: "Definitions") -> "Definitions":
        """Merges multiple Definitions objects into a single Definitions object.

        The returned Definitions object has the union of all the definitions in the input
        Definitions objects.

        Raises an error if the Definitions objects to be merged contain conflicting values for the
        same resource key or logger key, or if they have different executors defined.

        Examples:
            .. code-block:: python

                import submodule1
                import submodule2

                defs = Definitions.merge(submodule1.defs, submodule2.defs)

        Returns:
            Definitions: The merged definitions.
        """
        check.sequence_param(def_sets, "def_sets", of_type=Definitions)
        assets = []
        schedules = []
        sensors = []
        jobs = []
        asset_checks = []
        metadata = {}
        component_tree = None

        resources = {}
        resource_key_indexes: dict[str, int] = {}
        loggers = {}
        logger_key_indexes: dict[str, int] = {}
        executor = None
        executor_index: int | None = None

        for i, def_set in enumerate(def_sets):
            assets.extend(def_set.assets or [])
            asset_checks.extend(def_set.asset_checks or [])
            schedules.extend(def_set.schedules or [])
            sensors.extend(def_set.sensors or [])
            jobs.extend(def_set.jobs or [])
            metadata.update(def_set.metadata)

            for resource_key, resource_value in (def_set.resources or {}).items():
                if resource_key in resources and resources[resource_key] is not resource_value:
                    raise DagsterInvariantViolationError(
                        f"Definitions objects {resource_key_indexes[resource_key]} and {i} have "
                        f"different resources with same key '{resource_key}'"
                    )
                resources[resource_key] = resource_value
                resource_key_indexes[resource_key] = i

            for logger_key, logger_value in (def_set.loggers or {}).items():
                if logger_key in loggers and loggers[logger_key] is not logger_value:
                    raise DagsterInvariantViolationError(
                        f"Definitions objects {logger_key_indexes[logger_key]} and {i} have "
                        f"different loggers with same key '{logger_key}'"
                    )
                loggers[logger_key] = logger_value
                logger_key_indexes[logger_key] = i

            if def_set.executor is not None:
                if executor is not None and executor is not def_set.executor:
                    raise DagsterInvariantViolationError(
                        f"Definitions objects {executor_index} and {i} both have an executor"
                    )

                executor = def_set.executor
                executor_index = i

            if def_set.component_tree:
                if component_tree is not None:
                    raise DagsterInvariantViolationError(
                        "Can not merge Definitions that both contain component_tree."
                    )

                component_tree = def_set.component_tree

        return Definitions(
            assets=assets,
            schedules=schedules,
            sensors=sensors,
            jobs=jobs,
            resources=resources,
            executor=executor,
            loggers=loggers,
            asset_checks=asset_checks,
            metadata=metadata,
            component_tree=component_tree,
        )

    @public
    @deprecated(
        breaking_version="1.11",
        additional_warn_text="Use resolve_all_asset_specs instead",
        subject="get_all_asset_specs",
    )
    def get_all_asset_specs(self) -> Sequence[AssetSpec]:
        """Returns an AssetSpec object for AssetsDefinitions and AssetSpec passed directly to the Definitions object."""
        return self.resolve_all_asset_specs()

    @public
    def resolve_all_asset_specs(self) -> Sequence[AssetSpec]:
        """Returns an AssetSpec object for every asset contained inside the resolved Definitions object."""
        asset_graph = self.resolve_asset_graph()
        return [asset_node.to_asset_spec() for asset_node in asset_graph.asset_nodes]

    @public
    def resolve_all_asset_keys(self) -> Sequence[AssetKey]:
        """Returns an AssetKey object for every asset contained inside the resolved Definitions object."""
        return [spec.key for spec in self.resolve_all_asset_specs()]

    @preview
    def with_reconstruction_metadata(self, reconstruction_metadata: Mapping[str, str]) -> Self:
        """Add reconstruction metadata to the Definitions object. This is typically used to cache data
        loaded from some external API that is computed during initialization of a code server.
        The cached data is then made available on the DefinitionsLoadContext during
        reconstruction of the same project context (such as a run worker), allowing use of the
        cached data to avoid additional external API queries. Values are expected to be serialized
        in advance and must be strings.
        """
        check.mapping_param(reconstruction_metadata, "reconstruction_metadata", key_type=str)
        for k, v in reconstruction_metadata.items():
            if not isinstance(v, str):
                raise DagsterInvariantViolationError(
                    f"Reconstruction metadata values must be strings. State-representing values are"
                    f" expected to be serialized before being passed as reconstruction metadata."
                    f" Got for key {k}:\n\n{v}"
                )
        normalized_metadata = {
            k: CodeLocationReconstructionMetadataValue(v)
            for k, v in reconstruction_metadata.items()
        }
        return copy(
            self,
            metadata={
                **(self.metadata or {}),
                **normalized_metadata,
            },
        )

    @public
    @preview
    def map_asset_specs(
        self,
        *,
        func: Callable[[AssetSpec], AssetSpec],
        selection: CoercibleToAssetSelection | None = None,
    ) -> "Definitions":
        """Map a function over the included AssetSpecs or AssetsDefinitions in this Definitions object, replacing specs in the sequence
        or specs in an AssetsDefinitions with the result of the function.

        Args:
            func (Callable[[AssetSpec], AssetSpec]): The function to apply to each AssetSpec.
            selection (Optional[Union[str, Sequence[str], Sequence[AssetKey], Sequence[Union[AssetsDefinition, SourceAsset]], AssetSelection]]): An asset selection to narrow down the set of assets to apply the function to. If not provided, applies to all assets.

        Returns:
            Definitions: A Definitions object where the AssetSpecs have been replaced with the result of the function where the selection applies.

        Examples:
            .. code-block:: python

                import dagster as dg

                my_spec = dg.AssetSpec("asset1")

                @dg.asset
                def asset1(_): ...

                @dg.asset
                def asset2(_): ...

                defs = Definitions(
                    assets=[asset1, asset2]
                )

                # Applies to asset1 and asset2
                mapped_defs = defs.map_asset_specs(
                    func=lambda s: s.merge_attributes(metadata={"new_key": "new_value"}),
                )

        """
        check.invariant(
            selection is None,
            "The selection parameter is no longer supported for map_asset_specs, Please use map_resolved_asset_specs instead",
        )

        return self.map_resolved_asset_specs(func=func, selection=None)

    @public
    @preview
    def map_resolved_asset_specs(
        self,
        *,
        func: Callable[[AssetSpec], AssetSpec],
        selection: CoercibleToAssetSelection | None = None,
    ) -> "Definitions":
        """Map a function over the included AssetSpecs or AssetsDefinitions in this Definitions object, replacing specs in the sequence.

        See map_asset_specs for more details.

        Supports selection and therefore requires resolving the Definitions object to a RepositoryDefinition when there is a selection.

        Examples:
            .. code-block:: python

                import dagster as dg

                my_spec = dg.AssetSpec("asset1")

                @dg.asset
                def asset1(_): ...

                @dg.asset
                def asset2(_): ...

                # Applies only to asset1
                mapped_defs = defs.map_resolved_asset_specs(
                    func=lambda s: s.replace_attributes(metadata={"new_key": "new_value"}),
                    selection="asset1",
                )
        """
        non_spec_asset_types = {
            type(d) for d in self.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))
        }

        if non_spec_asset_types:
            raise DagsterInvariantViolationError(
                "Can only map over AssetSpec or AssetsDefinition objects. "
                "Received objects of types: "
                f"{non_spec_asset_types}."
            )

        return self.permissive_map_resolved_asset_specs(
            func=func,
            selection=selection,
        )

    def permissive_map_resolved_asset_specs(
        self,
        func: Callable[[AssetSpec], AssetSpec],
        selection: CoercibleToAssetSelection | None,
    ) -> "Definitions":
        """This is a permissive version of map_resolved_asset_specs that allows for non-spec asset types, i.e. SourceAssets and CacheableAssetsDefinitions."""
        target_keys = None
        if selection:
            if isinstance(selection, str):
                selection = AssetSelection.from_string(selection, include_sources=True)
            else:
                selection = AssetSelection.from_coercible(selection)
            target_keys = selection.resolve(self.resolve_asset_graph())
        mappable = iter(
            d for d in self.assets or [] if isinstance(d, (AssetsDefinition, AssetSpec))
        )
        mapped_assets = map_asset_specs(
            lambda spec: func(spec) if (target_keys is None or spec.key in target_keys) else spec,
            mappable,
        )

        assets = [
            *mapped_assets,
            *[d for d in self.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))],
        ]
        return replace(self, assets=assets)

    def with_resources(self, resources: Mapping[str, Any] | None) -> "Definitions":
        return Definitions.merge(self, Definitions(resources=resources)) if resources else self

    def has_resolved_repository_def(self) -> bool:
        return len(get_cached_method_cache(self, self.get_repository_def.__name__)) > 0

    def with_definition_metadata_update(
        self, update: Callable[[RawMetadataMapping], RawMetadataMapping]
    ):
        """Run a provided update function on every contained definition that supports it
        to updated its metadata. Return a new Definitions object containing the updated objects.
        """
        updated_jobs = _update_jobs_metadata(self.jobs, update)
        updated_schedules = _update_schedules_metadata(self.schedules, update, updated_jobs)
        updated_sensors = _update_sensors_metadata(self.sensors, update, updated_jobs)
        updated_assets = _update_assets_metadata(self.assets, update)
        updated_asset_checks = _update_checks_metadata(self.asset_checks, update)

        return replace(
            self,
            jobs=updated_jobs.values(),
            schedules=updated_schedules,
            sensors=updated_sensors,
            assets=updated_assets,
            asset_checks=updated_asset_checks,
        )


def _update_assets_metadata(
    assets: TAssets,
    update: Callable[[RawMetadataMapping], RawMetadataMapping],
) -> TAssets:
    if not assets:
        return assets

    updated_assets = []
    for asset in assets:
        if isinstance(asset, AssetsDefinition):
            updated_assets.append(_update_assets_def_metadata(asset, update))
        elif isinstance(asset, AssetSpec):
            updated_assets.append(asset.replace_attributes(metadata=update(asset.metadata)))
        elif isinstance(asset, (SourceAsset, CacheableAssetsDefinition)):
            # these types are deprecated and do not support metadata updates, ignore
            updated_assets.append(asset)
        else:
            check.assert_never(asset)
    return updated_assets


def _update_schedules_metadata(
    schedules: TSchedules,
    update: Callable[[RawMetadataMapping], RawMetadataMapping],
    updated_jobs: Mapping[int, TJob],
) -> TSchedules:
    if not schedules:
        return schedules

    updated_schedules = []
    for schedule in schedules:
        if isinstance(schedule, ScheduleDefinition):
            # updated schedule
            new_attrs: dict[str, Any] = {"metadata": update(schedule.metadata)}

            if schedule.has_job:
                # use the already updated job if possible to ensure obj equality
                if id(schedule.job) in updated_jobs:
                    new_attrs["job"] = updated_jobs[id(schedule.job)]

                else:  # otherwise, update the job metadata too
                    new_attrs["job"] = schedule.job.with_metadata(
                        update(schedule.job.metadata or {})
                    )

            updated_schedules.append(schedule.with_attributes(**new_attrs))

        elif isinstance(schedule, UnresolvedPartitionedAssetScheduleDefinition):
            updated_schedules.append(schedule.with_metadata(update(schedule.metadata or {})))
        else:
            check.assert_never(schedule)

    return updated_schedules


def _update_sensors_metadata(
    sensors: TSensors,
    update: Callable[[RawMetadataMapping], RawMetadataMapping],
    updated_jobs: Mapping[int, TJob],
) -> TSensors:
    if not sensors:
        return sensors

    updated_sensors = []
    for sensor in sensors:
        if isinstance(sensor, SensorDefinition):
            new_attrs: dict[str, Any] = {"metadata": update(sensor.metadata)}

            if sensor.has_jobs:
                new_sensor_jobs = []
                for sensor_job in sensor.jobs:
                    if isinstance(sensor_job, (JobDefinition, UnresolvedAssetJobDefinition)):
                        # use the already updated job if possible to ensure obj equality
                        if id(sensor_job) in updated_jobs:
                            new_sensor_jobs.append(updated_jobs[id(sensor_job)])
                        else:  # otherwise, update the job metadata too
                            new_sensor_jobs.append(
                                sensor_job.with_metadata(update(sensor_job.metadata or {}))
                            )
                    else:
                        new_sensor_jobs.append(sensor_job)  # other types are not updated
                new_attrs["jobs"] = new_sensor_jobs

            updated_sensors.append(sensor.with_attributes(**new_attrs))
        else:
            check.assert_never(sensor)

    return updated_sensors


def _update_jobs_metadata(
    jobs: TJobs,
    update: Callable[[RawMetadataMapping], RawMetadataMapping],
) -> Mapping[int, TJob]:
    if not jobs:
        return {}

    updated_jobs = {}
    for job in jobs:
        if isinstance(job, (JobDefinition, UnresolvedAssetJobDefinition)):
            updated_jobs[id(job)] = job.with_metadata(update(job.metadata or {}))
        else:
            check.assert_never(job)

    return updated_jobs


def _update_checks_metadata(
    asset_checks: TAssetChecks,
    update: Callable[[RawMetadataMapping], RawMetadataMapping],
) -> TAssetChecks:
    if not asset_checks:
        return asset_checks

    updated_checks = []
    for asset_check in asset_checks:
        if isinstance(asset_check, AssetsDefinition):
            updated_checks.append(_update_assets_def_metadata(asset_check, update))
        else:
            check.assert_never(asset_check)

    return updated_checks


def _update_assets_def_metadata(
    assets_def: AssetsDefinition,
    update: Callable[[RawMetadataMapping], RawMetadataMapping],
) -> AssetsDefinition:
    return assets_def.with_attributes(
        metadata_by_key={
            **{key: update(metadata) for key, metadata in assets_def.metadata_by_key.items()},
            **{c.key: update(c.metadata) for c in assets_def.check_specs},
        }
    )



from abc import ABC, abstractmethod

from dagster._annotations import public
from dagster._core.instance import MayHaveInstanceWeakref, T_DagsterInstance
from dagster._core.storage.event_log.base import EventLogStorage
from dagster._core.storage.runs.base import RunStorage
from dagster._core.storage.schedules.base import ScheduleStorage


@public
class DagsterStorage(ABC, MayHaveInstanceWeakref[T_DagsterInstance]):
    """Abstract base class for Dagster persistent storage, for reading and writing data for runs,
    events, and schedule/sensor state.

    Users should not directly instantiate concrete subclasses of this class; they are instantiated
    by internal machinery when ``dagster-webserver`` and ``dagster-daemon`` load, based on the values in the
    ``dagster.yaml`` file in ``$DAGSTER_HOME``. Configuration of concrete subclasses of this class
    should be done by setting values in that file.
    """

    @property
    @abstractmethod
    def event_log_storage(self) -> EventLogStorage[T_DagsterInstance]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def run_storage(self) -> RunStorage[T_DagsterInstance]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def schedule_storage(self) -> ScheduleStorage[T_DagsterInstance]:
        raise NotImplementedError()



import logging
import os
import warnings
from collections.abc import Mapping, Sequence, Set
from typing import TYPE_CHECKING, Any, Optional, cast, overload

import dagster._check as check
from dagster._core.definitions.asset_checks.asset_check_evaluation import (
    AssetCheckEvaluationPlanned,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.partitions.utils.time_window import TimeWindow
from dagster._core.errors import (
    DagsterInvalidSubsetError,
    DagsterInvariantViolationError,
    DagsterRunAlreadyExists,
    DagsterRunConflict,
)
from dagster._core.instance.utils import (
    AIRFLOW_EXECUTION_DATE_STR,
    IS_AIRFLOW_INGEST_PIPELINE_STR,
    _check_run_equality,
    _format_field_diff,
)
from dagster._core.origin import JobPythonOrigin
from dagster._core.storage.dagster_run import (
    DagsterRun,
    DagsterRunStatus,
    assets_are_externally_managed,
)
from dagster._core.storage.tags import (
    ASSET_PARTITION_RANGE_END_TAG,
    ASSET_PARTITION_RANGE_START_TAG,
    ASSET_RESUME_RETRY_TAG,
    BACKFILL_ID_TAG,
    BACKFILL_TAGS,
    CODE_LOCATION_TAG,
    PARENT_RUN_ID_TAG,
    PARTITION_NAME_TAG,
    RESUME_RETRY_TAG,
    ROOT_RUN_ID_TAG,
    TAGS_TO_MAYBE_OMIT_ON_RETRY,
)
from dagster._time import get_current_datetime
from dagster._utils import is_uuid
from dagster._utils.merger import merge_dicts
from dagster._utils.warnings import disable_dagster_warnings

if TYPE_CHECKING:
    from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
    from dagster._core.definitions.assets.graph.base_asset_graph import (
        AssetCheckNode,
        BaseAssetGraph,
        BaseAssetNode,
        BaseEntityNode,
    )
    from dagster._core.definitions.job_definition import JobDefinition
    from dagster._core.definitions.partitions.definition import PartitionsDefinition
    from dagster._core.definitions.repository_definition.repository_definition import (
        RepositoryLoadData,
    )
    from dagster._core.definitions.utils import EntityKey
    from dagster._core.events import DagsterEvent
    from dagster._core.execution.plan.plan import ExecutionPlan
    from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
    from dagster._core.instance.instance import DagsterInstance
    from dagster._core.remote_origin import RemoteJobOrigin
    from dagster._core.remote_representation.code_location import CodeLocation
    from dagster._core.remote_representation.external import RemoteJob
    from dagster._core.snap import ExecutionPlanSnapshot, JobSnap
    from dagster._core.snap.execution_plan_snapshot import ExecutionStepSnap
    from dagster._core.workspace.context import BaseWorkspaceRequestContext


class RunDomain:
    """Domain object encapsulating run-related operations.

    This class holds a reference to a DagsterInstance and provides methods
    for creating, managing, and querying runs.
    """

    def __init__(self, instance: "DagsterInstance") -> None:
        self._instance = instance

    def create_run(
        self,
        *,
        job_name: str,
        run_id: str | None,
        run_config: Mapping[str, object] | None,
        status: DagsterRunStatus | None,
        tags: Mapping[str, Any] | None,
        root_run_id: str | None,
        parent_run_id: str | None,
        step_keys_to_execute: Sequence[str] | None,
        execution_plan_snapshot: Optional["ExecutionPlanSnapshot"],
        job_snapshot: Optional["JobSnap"],
        parent_job_snapshot: Optional["JobSnap"],
        asset_selection: Set[AssetKey] | None,
        asset_check_selection: Set["AssetCheckKey"] | None,
        resolved_op_selection: Set[str] | None,
        op_selection: Sequence[str] | None,
        remote_job_origin: Optional["RemoteJobOrigin"],
        job_code_origin: JobPythonOrigin | None,
        asset_graph: "BaseAssetGraph",
    ) -> DagsterRun:
        """Create a run with the given parameters."""
        from dagster._core.definitions.asset_key import AssetCheckKey
        from dagster._core.definitions.assets.graph.remote_asset_graph import RemoteAssetGraph
        from dagster._core.definitions.partitions.context import partition_loading_context
        from dagster._core.remote_origin import RemoteJobOrigin
        from dagster._core.snap import ExecutionPlanSnapshot, JobSnap
        from dagster._utils.tags import normalize_tags

        check.str_param(job_name, "job_name")
        check.opt_str_param(
            run_id, "run_id"
        )  # will be assigned to make_new_run_id() lower in callstack
        check.opt_mapping_param(run_config, "run_config", key_type=str)

        check.opt_inst_param(status, "status", DagsterRunStatus)
        check.opt_mapping_param(tags, "tags", key_type=str)

        with disable_dagster_warnings():
            validated_tags = normalize_tags(tags)

        check.opt_str_param(root_run_id, "root_run_id")
        check.opt_str_param(parent_run_id, "parent_run_id")

        # If step_keys_to_execute is None, then everything is executed.  In some cases callers
        # are still exploding and sending the full list of step keys even though that is
        # unnecessary.

        check.opt_sequence_param(step_keys_to_execute, "step_keys_to_execute")
        check.opt_inst_param(
            execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot
        )

        if run_id and not is_uuid(run_id):
            check.failed(f"run_id must be a valid UUID. Got {run_id}")

        if root_run_id or parent_run_id:
            check.invariant(
                root_run_id and parent_run_id,
                "If root_run_id or parent_run_id is passed, this is a re-execution scenario and"
                " root_run_id and parent_run_id must both be passed.",
            )

        # The job_snapshot should always be set in production scenarios. In tests
        # we have sometimes omitted it out of convenience.

        check.opt_inst_param(job_snapshot, "job_snapshot", JobSnap)
        check.opt_inst_param(parent_job_snapshot, "parent_job_snapshot", JobSnap)

        if parent_job_snapshot:
            check.invariant(
                job_snapshot,
                "If parent_job_snapshot is set, job_snapshot should also be.",
            )

        # op_selection is a sequence of selection queries assigned by the user.
        # *Most* callers expand the op_selection into an explicit set of
        # resolved_op_selection via accessing remote_job.resolved_op_selection
        # but not all do. Some (launch execution mutation in graphql and backfill run
        # creation, for example) actually pass the solid *selection* into the
        # resolved_op_selection parameter, but just as a frozen set, rather than
        # fully resolving the selection, as the daemon launchers do. Given the
        # state of callers we just check to ensure that the arguments are well-formed.
        #
        # asset_selection adds another dimension to this lovely dance. op_selection
        # and asset_selection are mutually exclusive and should never both be set.
        # This is invariant is checked in a sporadic fashion around
        # the codebase, but is never enforced in a typed fashion.
        #
        # Additionally, the way that callsites currently behave *if* asset selection
        # is set (i.e., not None) then *neither* op_selection *nor*
        # resolved_op_selection is passed. In the asset selection case resolving
        # the set of assets into the canonical resolved_op_selection is done in
        # the user process, and the exact resolution is never persisted in the run.
        # We are asserting that invariant here to maintain that behavior.
        #
        # Finally, asset_check_selection can be passed along with asset_selection. It
        # is mutually exclusive with op_selection and resolved_op_selection. A `None`
        # value will include any asset checks that target selected assets. An empty set
        # will include no asset checks.

        check.opt_set_param(resolved_op_selection, "resolved_op_selection", of_type=str)
        check.opt_sequence_param(op_selection, "op_selection", of_type=str)
        check.opt_set_param(asset_selection, "asset_selection", of_type=AssetKey)
        check.opt_set_param(asset_check_selection, "asset_check_selection", of_type=AssetCheckKey)

        # asset_selection will always be None on an op job, but asset_check_selection may be
        # None or []. This is because [] and None are different for asset checks: None means
        # include all asset checks on selected assets, while [] means include no asset checks.
        # In an op job (which has no asset checks), these two are equivalent.
        if asset_selection is not None or asset_check_selection:
            check.invariant(
                op_selection is None,
                "Cannot pass op_selection with either of asset_selection or asset_check_selection",
            )

            check.invariant(
                resolved_op_selection is None,
                "Cannot pass resolved_op_selection with either of asset_selection or"
                " asset_check_selection",
            )

        # The "python origin" arguments exist so a job can be reconstructed in memory
        # after a DagsterRun has been fetched from the database.
        #
        # There are cases (notably in _logged_execute_job with Reconstructable jobs)
        # where job_code_origin and is not. In some cloud test cases only
        # remote_job_origin is passed But they are almost always passed together.
        # If these are not set the created run will never be able to be relaunched from
        # the information just in the run or in another process.

        check.opt_inst_param(remote_job_origin, "remote_job_origin", RemoteJobOrigin)
        check.opt_inst_param(job_code_origin, "job_code_origin", JobPythonOrigin)

        if isinstance(asset_graph, RemoteAssetGraph):
            check.invariant(
                remote_job_origin is not None,
                "must include a remote job origin when creating a run using a remote asset graph",
            )

        # Add the code location tag to the validated tags
        # This is useful for filtering runs by code location in the web server
        # or control concurrency by code location with tag concurrency limits.
        if remote_job_origin:
            validated_tags = {
                **validated_tags,
                CODE_LOCATION_TAG: remote_job_origin.repository_origin.code_location_origin.location_name,
            }

        dagster_run = self.construct_run_with_snapshots(
            job_name=job_name,
            run_id=run_id,  # type: ignore  # (possible none)
            run_config=run_config,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
            op_selection=op_selection,
            resolved_op_selection=resolved_op_selection,
            step_keys_to_execute=step_keys_to_execute,
            status=status,
            tags=validated_tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            job_snapshot=job_snapshot,
            execution_plan_snapshot=execution_plan_snapshot,
            parent_job_snapshot=parent_job_snapshot,
            remote_job_origin=remote_job_origin,
            job_code_origin=job_code_origin,
            asset_graph=asset_graph,
        )

        dagster_run = self._instance.run_storage.add_run(dagster_run)

        if execution_plan_snapshot and not assets_are_externally_managed(dagster_run):
            with partition_loading_context(dynamic_partitions_store=self._instance):
                self._log_asset_planned_events(dagster_run, execution_plan_snapshot, asset_graph)

        return dagster_run

    def construct_run_with_snapshots(
        self,
        job_name: str,
        run_id: str,
        run_config: Mapping[str, object] | None,
        resolved_op_selection: Set[str] | None,
        step_keys_to_execute: Sequence[str] | None,
        status: DagsterRunStatus | None,
        tags: Mapping[str, str],
        root_run_id: str | None,
        parent_run_id: str | None,
        job_snapshot: Optional["JobSnap"],
        execution_plan_snapshot: Optional["ExecutionPlanSnapshot"],
        parent_job_snapshot: Optional["JobSnap"],
        asset_selection: Set[AssetKey] | None = None,
        asset_check_selection: Set["AssetCheckKey"] | None = None,
        op_selection: Sequence[str] | None = None,
        remote_job_origin: Optional["RemoteJobOrigin"] = None,
        job_code_origin: JobPythonOrigin | None = None,
        asset_graph: Optional["BaseAssetGraph[BaseAssetNode]"] = None,
    ) -> DagsterRun:
        """Heavy run construction logic moved from DagsterInstance."""
        # https://github.com/dagster-io/dagster/issues/2403
        if tags and IS_AIRFLOW_INGEST_PIPELINE_STR in tags:
            if AIRFLOW_EXECUTION_DATE_STR not in tags:
                tags = {
                    **tags,
                    AIRFLOW_EXECUTION_DATE_STR: get_current_datetime().isoformat(),
                }

        check.invariant(
            not (not job_snapshot and execution_plan_snapshot),
            "It is illegal to have an execution plan snapshot and not have a pipeline snapshot."
            " It is possible to have no execution plan snapshot since we persist runs that do"
            " not successfully compile execution plans in the scheduled case.",
        )

        job_snapshot_id = (
            self.ensure_persisted_job_snapshot(job_snapshot, parent_job_snapshot)
            if job_snapshot
            else None
        )
        partitions_definition = None

        # ensure that all asset outputs list their execution type, even if the snapshot was
        # created on an older version before it was being set
        if execution_plan_snapshot and asset_graph:
            adjusted_steps = []
            for step in execution_plan_snapshot.steps:
                adjusted_outputs = []
                for output in step.outputs:
                    asset_key = output.properties.asset_key if output.properties else None
                    adjusted_output = output

                    if asset_key:
                        asset_node = self._get_repo_scoped_entity_node(
                            asset_key, asset_graph, remote_job_origin
                        )
                        if asset_node:
                            partitions_definition = asset_node.partitions_def

                            if (
                                partitions_definition is None
                                and asset_node.partitions_def is not None
                            ):
                                # this assumes that if one partitioned asset is in a run, all other partitioned
                                # assets in the run have the same partitions definition.
                                partitions_definition = asset_node.partitions_def

                            if (
                                output.properties is not None
                                and output.properties.asset_execution_type is None
                            ):
                                adjusted_output = output._replace(
                                    properties=output.properties._replace(
                                        asset_execution_type=asset_node.execution_type
                                    )
                                )

                    adjusted_outputs.append(adjusted_output)

                adjusted_steps.append(step._replace(outputs=adjusted_outputs))

            execution_plan_snapshot = execution_plan_snapshot._replace(steps=adjusted_steps)

        execution_plan_snapshot_id = (
            self.ensure_persisted_execution_plan_snapshot(
                execution_plan_snapshot, job_snapshot_id, step_keys_to_execute
            )
            if execution_plan_snapshot and job_snapshot_id
            else None
        )

        if execution_plan_snapshot:
            from dagster._core.op_concurrency_limits_counter import (
                compute_run_op_concurrency_info_for_snapshot,
            )

            run_op_concurrency = compute_run_op_concurrency_info_for_snapshot(
                execution_plan_snapshot
            )
        else:
            run_op_concurrency = None

        # Calculate partitions_subset for time window partitions
        partitions_subset = None
        if partitions_definition is not None:
            from dagster._core.definitions.partitions.definition import DynamicPartitionsDefinition
            from dagster._core.definitions.partitions.definition.time_window import (
                TimeWindowPartitionsDefinition,
            )
            from dagster._core.definitions.partitions.subset import KeyRangesPartitionsSubset
            from dagster._core.remote_representation.external_data import PartitionsSnap

            partition_tag = tags.get(PARTITION_NAME_TAG)
            partition_key_range = (
                PartitionKeyRange(
                    tags[ASSET_PARTITION_RANGE_START_TAG],
                    tags[ASSET_PARTITION_RANGE_END_TAG],
                )
                if (
                    tags.get(ASSET_PARTITION_RANGE_START_TAG) is not None
                    and tags.get(ASSET_PARTITION_RANGE_END_TAG) is not None
                )
                else None
            )
            if partition_tag is not None or partition_key_range is not None:
                if partition_tag is not None and partition_key_range is not None:
                    raise DagsterInvariantViolationError(
                        f"Cannot have {ASSET_PARTITION_RANGE_START_TAG} or"
                        f" {ASSET_PARTITION_RANGE_END_TAG} tags set along with"
                        f" {PARTITION_NAME_TAG}"
                    )
                if partition_tag is not None:
                    partition_key_range = PartitionKeyRange(partition_tag, partition_tag)
            if partition_key_range is not None:
                # only store certain subsets that can be represented compactly
                if isinstance(partitions_definition, TimeWindowPartitionsDefinition):
                    start_window = partitions_definition.time_window_for_partition_key(
                        partition_key_range.start
                    )
                    end_window = partitions_definition.time_window_for_partition_key(
                        partition_key_range.end
                    )
                    partitions_subset = partitions_definition.get_partition_subset_in_time_window(
                        TimeWindow(start_window.start, end_window.end)
                    ).to_serializable_subset()

                    # only store the subset of time window partitions, since those can be compressed efficiently
                elif (
                    isinstance(partitions_definition, DynamicPartitionsDefinition)
                    and partitions_definition.name is not None
                ):
                    partitions_subset = KeyRangesPartitionsSubset(
                        key_ranges=[partition_key_range],
                        partitions_snap=PartitionsSnap.from_def(partitions_definition),
                    )

        return DagsterRun(
            job_name=job_name,
            run_id=run_id,
            run_config=run_config,
            asset_selection=asset_selection,
            asset_check_selection=asset_check_selection,
            op_selection=op_selection,
            resolved_op_selection=resolved_op_selection,
            step_keys_to_execute=step_keys_to_execute,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            job_snapshot_id=job_snapshot_id,
            execution_plan_snapshot_id=execution_plan_snapshot_id,
            remote_job_origin=remote_job_origin,
            job_code_origin=job_code_origin,
            has_repository_load_data=execution_plan_snapshot is not None
            and execution_plan_snapshot.repository_load_data is not None,
            run_op_concurrency=run_op_concurrency,
            partitions_subset=partitions_subset,
        )

    def create_reexecuted_run(
        self,
        *,
        request_context: "BaseWorkspaceRequestContext",
        parent_run: DagsterRun,
        code_location: "CodeLocation",
        remote_job: "RemoteJob",
        strategy: "ReexecutionStrategy",
        extra_tags: Mapping[str, Any] | None = None,
        run_config: Mapping[str, Any] | None = None,
        use_parent_run_tags: bool = False,
    ) -> DagsterRun:
        """Reexecution logic moved from DagsterInstance."""
        from dagster._core.execution.backfill import BulkActionStatus
        from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
        from dagster._core.execution.plan.state import KnownExecutionState
        from dagster._core.remote_representation.code_location import CodeLocation
        from dagster._core.remote_representation.external import RemoteJob

        check.inst_param(parent_run, "parent_run", DagsterRun)
        check.inst_param(code_location, "code_location", CodeLocation)
        check.inst_param(remote_job, "remote_job", RemoteJob)
        check.inst_param(strategy, "strategy", ReexecutionStrategy)
        check.opt_mapping_param(extra_tags, "extra_tags", key_type=str)
        check.opt_mapping_param(run_config, "run_config", key_type=str)

        check.bool_param(use_parent_run_tags, "use_parent_run_tags")

        root_run_id = parent_run.root_run_id or parent_run.run_id
        parent_run_id = parent_run.run_id

        # these can differ from remote_job.tags if tags were added at launch time
        parent_run_tags_to_include = {}
        if use_parent_run_tags:
            parent_run_tags_to_include = {
                key: val
                for key, val in parent_run.tags.items()
                if key not in TAGS_TO_MAYBE_OMIT_ON_RETRY
            }
            # condition to determine whether to include BACKFILL_ID_TAG, PARENT_BACKFILL_ID_TAG,
            # ROOT_BACKFILL_ID_TAG on retried run
            if parent_run.tags.get(BACKFILL_ID_TAG) is not None:
                # if the run was part of a backfill and the backfill is complete, we do not want the
                # retry to be considered part of the backfill, so remove all backfill-related tags
                backfill = self._instance.get_backfill(parent_run.tags[BACKFILL_ID_TAG])
                if backfill and backfill.status == BulkActionStatus.REQUESTED:
                    for tag in BACKFILL_TAGS:
                        if parent_run.tags.get(tag) is not None:
                            parent_run_tags_to_include[tag] = parent_run.tags[tag]

        tags = merge_dicts(
            remote_job.tags,
            parent_run_tags_to_include,
            extra_tags or {},
            {
                PARENT_RUN_ID_TAG: parent_run_id,
                ROOT_RUN_ID_TAG: root_run_id,
            },
        )

        run_config = run_config if run_config is not None else parent_run.run_config

        if strategy == ReexecutionStrategy.FROM_FAILURE:
            (
                step_keys_to_execute,
                known_state,
            ) = KnownExecutionState.build_resume_retry_reexecution(
                self._instance,
                parent_run=parent_run,
            )
            tags[RESUME_RETRY_TAG] = "true"

            if not step_keys_to_execute:
                raise DagsterInvalidSubsetError("No steps needed to be retried in the failed run.")
        elif strategy == ReexecutionStrategy.FROM_ASSET_FAILURE:
            parent_snapshot_id = check.not_none(parent_run.execution_plan_snapshot_id)
            snapshot = self._instance.get_execution_plan_snapshot(parent_snapshot_id)
            skipped_asset_keys, skipped_asset_check_keys = self.get_keys_to_reexecute(
                parent_run_id, snapshot
            )

            if not skipped_asset_keys and not skipped_asset_check_keys:
                raise DagsterInvalidSubsetError(
                    "No assets or asset checks needed to be retried in the failed run."
                )

            remote_job = code_location.get_job(
                remote_job.get_subset_selector(
                    asset_selection=skipped_asset_keys,
                    asset_check_selection=skipped_asset_check_keys,
                )
            )
            step_keys_to_execute = None
            known_state = None
            tags[ASSET_RESUME_RETRY_TAG] = "true"
        elif strategy == ReexecutionStrategy.ALL_STEPS:
            step_keys_to_execute = None
            known_state = None
        else:
            raise DagsterInvariantViolationError(f"Unknown reexecution strategy: {strategy}")

        remote_execution_plan = code_location.get_execution_plan(
            remote_job,
            run_config,
            step_keys_to_execute=step_keys_to_execute,
            known_state=known_state,
            instance=self._instance,
        )

        return self.create_run(
            job_name=parent_run.job_name,
            run_id=None,
            run_config=run_config,
            resolved_op_selection=parent_run.resolved_op_selection,
            step_keys_to_execute=step_keys_to_execute,
            status=DagsterRunStatus.NOT_STARTED,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            job_snapshot=remote_job.job_snapshot,
            execution_plan_snapshot=remote_execution_plan.execution_plan_snapshot,
            parent_job_snapshot=remote_job.parent_job_snapshot,
            op_selection=parent_run.op_selection,
            asset_selection=remote_job.asset_selection,
            asset_check_selection=remote_job.asset_check_selection,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
            asset_graph=request_context.asset_graph,
        )

    def register_managed_run(
        self,
        job_name: str,
        run_id: str,
        run_config: Mapping[str, object] | None,
        resolved_op_selection: Set[str] | None,
        step_keys_to_execute: Sequence[str] | None,
        tags: Mapping[str, str],
        root_run_id: str | None,
        parent_run_id: str | None,
        job_snapshot: Optional["JobSnap"],
        execution_plan_snapshot: Optional["ExecutionPlanSnapshot"],
        parent_job_snapshot: Optional["JobSnap"],
        op_selection: Sequence[str] | None = None,
        job_code_origin: JobPythonOrigin | None = None,
    ) -> DagsterRun:
        """Managed run registration moved from DagsterInstance."""
        # The usage of this method is limited to dagster-airflow, specifically in Dagster
        # Operators that are executed in Airflow. Because a common workflow in Airflow is to
        # retry dags from arbitrary tasks, we need any node to be capable of creating a
        # DagsterRun.
        #
        # The try-except DagsterRunAlreadyExists block handles the race when multiple "root" tasks
        # simultaneously execute self._run_storage.add_run(dagster_run). When this happens, only
        # one task succeeds in creating the run, while the others get DagsterRunAlreadyExists
        # error; at this point, the failed tasks try again to fetch the existing run.
        # https://github.com/dagster-io/dagster/issues/2412

        dagster_run = self.construct_run_with_snapshots(
            job_name=job_name,
            run_id=run_id,
            run_config=run_config,
            op_selection=op_selection,
            resolved_op_selection=resolved_op_selection,
            step_keys_to_execute=step_keys_to_execute,
            status=DagsterRunStatus.MANAGED,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            job_snapshot=job_snapshot,
            execution_plan_snapshot=execution_plan_snapshot,
            parent_job_snapshot=parent_job_snapshot,
            job_code_origin=job_code_origin,
        )

        def get_run() -> DagsterRun:
            candidate_run = self._instance.get_run_by_id(dagster_run.run_id)

            field_diff = _check_run_equality(dagster_run, candidate_run)  # type: ignore  # (possible none)

            if field_diff:
                raise DagsterRunConflict(
                    f"Found conflicting existing run with same id {dagster_run.run_id}. Runs differ in:"
                    f"\n{_format_field_diff(field_diff)}",
                )
            return candidate_run  # type: ignore  # (possible none)

        if self._instance.has_run(dagster_run.run_id):
            return get_run()

        try:
            return self._instance.run_storage.add_run(dagster_run)
        except DagsterRunAlreadyExists:
            return get_run()

    def ensure_persisted_job_snapshot(
        self,
        job_snapshot: "JobSnap",
        parent_job_snapshot: "JobSnap | None",
    ) -> str:
        """Moved from DagsterInstance._ensure_persisted_job_snapshot."""
        from dagster._core.snap import JobSnap

        check.inst_param(job_snapshot, "job_snapshot", JobSnap)
        check.opt_inst_param(parent_job_snapshot, "parent_job_snapshot", JobSnap)

        if job_snapshot.lineage_snapshot and parent_job_snapshot:
            parent_snapshot_id = check.not_none(parent_job_snapshot).snapshot_id

            if job_snapshot.lineage_snapshot.parent_snapshot_id != parent_snapshot_id:
                warnings.warn(
                    f"Stored parent snapshot ID {parent_snapshot_id} did not match the parent snapshot ID {job_snapshot.lineage_snapshot.parent_snapshot_id} on the subsetted job"
                )

            if not self._instance.run_storage.has_job_snapshot(parent_snapshot_id):
                self._instance.run_storage.add_job_snapshot(check.not_none(parent_job_snapshot))

        job_snapshot_id = job_snapshot.snapshot_id
        if not self._instance.run_storage.has_job_snapshot(job_snapshot_id):
            returned_job_snapshot_id = self._instance.run_storage.add_job_snapshot(job_snapshot)
            check.invariant(job_snapshot_id == returned_job_snapshot_id)

        return job_snapshot_id

    def ensure_persisted_execution_plan_snapshot(
        self,
        execution_plan_snapshot: "ExecutionPlanSnapshot",
        job_snapshot_id: str,
        step_keys_to_execute: Sequence[str] | None,
    ) -> str:
        """Moved from DagsterInstance._ensure_persisted_execution_plan_snapshot."""
        from dagster._core.snap.execution_plan_snapshot import (
            ExecutionPlanSnapshot,
            create_execution_plan_snapshot_id,
        )

        check.inst_param(execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot)
        check.str_param(job_snapshot_id, "job_snapshot_id")
        check.opt_nullable_sequence_param(step_keys_to_execute, "step_keys_to_execute", of_type=str)

        check.invariant(
            execution_plan_snapshot.job_snapshot_id == job_snapshot_id,
            "Snapshot mismatch: Snapshot ID in execution plan snapshot is "
            f'"{execution_plan_snapshot.job_snapshot_id}" and snapshot_id created in memory is '
            f'"{job_snapshot_id}"',
        )

        execution_plan_snapshot_id = create_execution_plan_snapshot_id(execution_plan_snapshot)

        if not self._instance.run_storage.has_execution_plan_snapshot(execution_plan_snapshot_id):
            returned_execution_plan_snapshot_id = (
                self._instance.run_storage.add_execution_plan_snapshot(execution_plan_snapshot)
            )

            check.invariant(execution_plan_snapshot_id == returned_execution_plan_snapshot_id)

        return execution_plan_snapshot_id

    def get_keys_to_reexecute(
        self,
        run_id: str,
        execution_plan_snapshot: "ExecutionPlanSnapshot",
    ) -> tuple[Set["AssetKey"], Set["AssetCheckKey"]]:
        """For a given run_id, return the subset of asset keys and asset check keys that should be
        re-executed for a run when in the `FROM_ASSET_FAILURE` mode.

        An asset key will be included if it was planned but not materialized in the original run,
        or if any of its planned blocking asset checks were planned but not executed, or failed.

        An asset check key will be included if it was planned but not executed in the original run,
        or if it was associated with an asset that will be re-executed.
        """
        from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey
        from dagster._core.events import (
            AssetCheckEvaluation,
            DagsterEventType,
            StepMaterializationData,
        )

        # figure out the set of assets that were materialized and checks that successfully executed
        logs = self._instance.all_logs(
            run_id=run_id,
            of_type={
                DagsterEventType.ASSET_MATERIALIZATION,
                DagsterEventType.ASSET_CHECK_EVALUATION,
            },
        )
        executed_keys: set[EntityKey] = set()
        blocking_failure_keys: set[AssetKey] = set()
        for log in logs:
            event_data = log.dagster_event.event_specific_data if log.dagster_event else None
            if isinstance(event_data, StepMaterializationData):
                executed_keys.add(event_data.materialization.asset_key)
            elif isinstance(event_data, AssetCheckEvaluation):
                # blocking asset checks did not "successfully execute", so we keep track
                # of them and their associated assets
                if event_data.blocking and not event_data.passed:
                    blocking_failure_keys.add(event_data.asset_check_key.asset_key)
                else:
                    executed_keys.add(event_data.asset_check_key)

        # handled_keys is the set of keys that do not need to be re-executed
        to_not_reexecute = executed_keys - blocking_failure_keys

        # find the set of planned assets and checks
        to_reexecute: set[EntityKey] = set()
        for step in execution_plan_snapshot.steps:
            if step.key not in execution_plan_snapshot.step_keys_to_execute:
                continue
            to_reexecute_for_step = {
                key
                for key in step.entity_keys
                if key not in to_not_reexecute
                # we need to re-execute any asset check keys (blocking or otherwise) if the asset
                # has a failed blocking check.
                or (isinstance(key, AssetCheckKey) and key.asset_key in blocking_failure_keys)
            }
            if to_reexecute_for_step:
                # we need to include all keys that were marked as required on the step
                to_reexecute.update(to_reexecute_for_step | step.required_entity_keys)

        return (
            {key for key in to_reexecute if isinstance(key, AssetKey)},
            {key for key in to_reexecute if isinstance(key, AssetCheckKey)},
        )

    @overload
    def _get_repo_scoped_entity_node(
        self,
        key: AssetKey,
        asset_graph: "BaseAssetGraph",
        remote_job_origin: Optional["RemoteJobOrigin"] = None,
    ) -> Optional["BaseAssetNode"]: ...

    @overload
    def _get_repo_scoped_entity_node(
        self,
        key: "AssetCheckKey",
        asset_graph: "BaseAssetGraph",
        remote_job_origin: Optional["RemoteJobOrigin"] = None,
    ) -> Optional["AssetCheckNode"]: ...

    def _get_repo_scoped_entity_node(
        self,
        key: "EntityKey",
        asset_graph: "BaseAssetGraph",
        remote_job_origin: Optional["RemoteJobOrigin"] = None,
    ) -> Optional["BaseEntityNode"]:
        from dagster._core.definitions.assets.graph.remote_asset_graph import (
            RemoteWorkspaceAssetGraph,
        )

        # the asset graph may be workspace-scoped (to take better advantage of request context
        # caching) or repository-scoped. In most cases it will be a remote asset graph, but in
        # tests or programatically creating runs from python codes it may be a regular asset graph.
        # in all cases, return the BaseAssetNode for the supplied asset key if it exists.
        if isinstance(asset_graph, RemoteWorkspaceAssetGraph):
            return cast(
                "BaseEntityNode | None",
                asset_graph.get_repo_scoped_node(
                    key, check.not_none(remote_job_origin).repository_origin.get_selector()
                ),
            )

        if not asset_graph.has(key):
            return None

        return asset_graph.get(key)

    def _get_partitions_def(
        self,
        key: "EntityKey",
        asset_graph: "BaseAssetGraph",
        remote_job_origin: Optional["RemoteJobOrigin"],
        run: "DagsterRun",
    ) -> Optional["PartitionsDefinition"]:
        # don't fetch the partitions def if the run is not partitioned
        if not run.is_partitioned:
            return None
        entity_node = self._get_repo_scoped_entity_node(key, asset_graph, remote_job_origin)
        return entity_node.partitions_def if entity_node else None

    def _log_asset_planned_events(
        self,
        dagster_run: DagsterRun,
        execution_plan_snapshot: "ExecutionPlanSnapshot",
        asset_graph: "BaseAssetGraph",
    ) -> None:
        """Moved from DagsterInstance._log_asset_planned_events."""
        from dagster._core.events import (
            DagsterEvent,
            DagsterEventBatchMetadata,
            DagsterEventType,
            generate_event_batch_id,
        )

        job_name = dagster_run.job_name

        events = []

        for step in execution_plan_snapshot.steps:
            if step.key in execution_plan_snapshot.step_keys_to_execute:
                for output in step.outputs:
                    asset_key = check.not_none(output.properties).asset_key
                    if asset_key:
                        events.extend(
                            self.get_materialization_planned_events_for_asset(
                                dagster_run, asset_key, job_name, step, asset_graph
                            )
                        )

                    if check.not_none(output.properties).asset_check_key:
                        asset_check_key = check.not_none(
                            check.not_none(output.properties).asset_check_key
                        )
                        target_asset_key = asset_check_key.asset_key
                        check_name = asset_check_key.name

                        partitions_def = self._get_partitions_def(
                            asset_check_key, asset_graph, dagster_run.remote_job_origin, dagster_run
                        )
                        partitions_subset = dagster_run.get_resolved_partitions_subset_for_events(
                            partitions_def
                        )

                        event = DagsterEvent(
                            event_type_value=DagsterEventType.ASSET_CHECK_EVALUATION_PLANNED.value,
                            job_name=job_name,
                            message=(
                                f"{job_name} intends to execute asset check {check_name} on"
                                f" asset {target_asset_key.to_string()}"
                            ),
                            event_specific_data=AssetCheckEvaluationPlanned(
                                asset_key=target_asset_key,
                                check_name=check_name,
                                partitions_subset=partitions_subset,
                            ),
                            step_key=step.key,
                        )
                        events.append(event)

        batch_id = generate_event_batch_id()
        last_index = len(events) - 1

        batch_planned_events = not bool(os.getenv("DAGSTER_EMIT_PLANNED_EVENTS_INDIVIDUALLY"))

        if batch_planned_events:
            # sort events by asset key to ensure they are inserted in a consistent order
            # if they are batched
            events = sorted(
                events,
                key=lambda event: (
                    event.asset_key.to_db_string()
                    if event.asset_key
                    else event.asset_check_planned_data.asset_check_key.to_db_string()
                ),
            )

        for i, event in enumerate(events):
            batch_metadata = (
                DagsterEventBatchMetadata(batch_id, i == last_index)
                if batch_planned_events
                else None
            )
            self._instance.report_dagster_event(
                event, dagster_run.run_id, logging.DEBUG, batch_metadata=batch_metadata
            )

    def get_materialization_planned_events_for_asset(
        self,
        dagster_run: DagsterRun,
        asset_key: AssetKey,
        job_name: str,
        step: "ExecutionStepSnap",
        asset_graph: "BaseAssetGraph[BaseAssetNode]",
    ) -> Sequence["DagsterEvent"]:
        """Moved from DagsterInstance._log_materialization_planned_event_for_asset."""
        from dagster._core.events import AssetMaterializationPlannedData, DagsterEvent

        events = []

        partitions_def = self._get_partitions_def(
            asset_key, asset_graph, dagster_run.remote_job_origin, dagster_run
        )

        partitions_subset = dagster_run.get_resolved_partitions_subset_for_events(partitions_def)
        if partitions_subset is None:
            materialization_planned = DagsterEvent.build_asset_materialization_planned_event(
                job_name,
                step.key,
                AssetMaterializationPlannedData(asset_key, partition=None, partitions_subset=None),
            )
            events.append(materialization_planned)
        elif self._instance.event_log_storage.supports_partition_subset_in_asset_materialization_planned_events:
            materialization_planned = DagsterEvent.build_asset_materialization_planned_event(
                job_name,
                step.key,
                AssetMaterializationPlannedData(
                    asset_key, partition=None, partitions_subset=partitions_subset
                ),
            )
            events.append(materialization_planned)
        else:
            for partition_key in partitions_subset.get_partition_keys():
                materialization_planned = DagsterEvent.build_asset_materialization_planned_event(
                    job_name,
                    step.key,
                    AssetMaterializationPlannedData(
                        asset_key,
                        partition=partition_key,
                        partitions_subset=None,
                    ),
                )
                events.append(materialization_planned)

        return events

    def create_run_for_job(
        self,
        job_def: "JobDefinition",
        execution_plan: Optional["ExecutionPlan"] = None,
        run_id: str | None = None,
        run_config: Mapping[str, object] | None = None,
        resolved_op_selection: Set[str] | None = None,
        status: DagsterRunStatus | None = None,
        tags: Mapping[str, str] | None = None,
        root_run_id: str | None = None,
        parent_run_id: str | None = None,
        op_selection: Sequence[str] | None = None,
        asset_selection: Set[AssetKey] | None = None,
        remote_job_origin: Optional["RemoteJobOrigin"] = None,
        job_code_origin: JobPythonOrigin | None = None,
        repository_load_data: Optional["RepositoryLoadData"] = None,
    ) -> DagsterRun:
        """Create run for job - moved from DagsterInstance.create_run_for_job()."""
        from dagster._core.definitions.job_definition import JobDefinition
        from dagster._core.execution.api import create_execution_plan
        from dagster._core.execution.plan.plan import ExecutionPlan
        from dagster._core.snap import snapshot_from_execution_plan

        check.inst_param(job_def, "pipeline_def", JobDefinition)
        check.opt_inst_param(execution_plan, "execution_plan", ExecutionPlan)
        # note that op_selection is required to execute the solid subset, which is the
        # frozenset version of the previous solid_subset.
        # op_selection is not required and will not be converted to op_selection here.
        # i.e. this function doesn't handle solid queries.
        # op_selection is only used to pass the user queries further down.
        check.opt_set_param(resolved_op_selection, "resolved_op_selection", of_type=str)
        check.opt_list_param(op_selection, "op_selection", of_type=str)
        check.opt_set_param(asset_selection, "asset_selection", of_type=AssetKey)

        # op_selection never provided
        if asset_selection or op_selection:
            # for cases when `create_run_for_pipeline` is directly called
            job_def = job_def.get_subset(
                asset_selection=asset_selection,
                op_selection=op_selection,
            )

        if not execution_plan:
            execution_plan = create_execution_plan(
                job=job_def,
                run_config=run_config,
                instance_ref=self._instance.get_ref() if self._instance.is_persistent else None,
                tags=tags,
                repository_load_data=repository_load_data,
            )

        return self.create_run(
            job_name=job_def.name,
            run_id=run_id,
            run_config=run_config,
            op_selection=op_selection,
            asset_selection=asset_selection,
            asset_check_selection=None,
            resolved_op_selection=resolved_op_selection,
            step_keys_to_execute=execution_plan.step_keys_to_execute,
            status=status,
            tags=tags,
            root_run_id=root_run_id,
            parent_run_id=parent_run_id,
            job_snapshot=job_def.get_job_snapshot(),
            execution_plan_snapshot=snapshot_from_execution_plan(
                execution_plan,
                job_def.get_job_snapshot_id(),
            ),
            parent_job_snapshot=job_def.get_parent_job_snapshot(),
            remote_job_origin=remote_job_origin,
            job_code_origin=job_code_origin,
            asset_graph=job_def.asset_layer.asset_graph,
        )



from dagster.components.testing.deprecated.utils import (
    get_all_components_defs_within_project as get_all_components_defs_within_project,
    get_component_defs_within_project as get_component_defs_within_project,
    get_module_path as get_module_path,
    scaffold_defs_sandbox as scaffold_defs_sandbox,
)
from dagster.components.testing.utils import (
    DefsFolderSandbox as DefsFolderSandbox,
    copy_code_to_file as copy_code_to_file,
    create_defs_folder_sandbox as create_defs_folder_sandbox,
    get_original_module_name as get_original_module_name,
    random_importable_name as random_importable_name,
)



from collections.abc import Mapping, Sequence
from typing import NamedTuple, cast

import dagster._check as check
from dagster._annotations import beta, public
from dagster._core.definitions.partitions.definition.multi import MultiPartitionsDefinition
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.mapping.multi.base import (
    BaseMultiPartitionMapping,
    DimensionDependency,
    DimensionPartitionMapping,
)
from dagster._core.definitions.partitions.mapping.partition_mapping import PartitionMapping
from dagster._serdes import whitelist_for_serdes


@beta
@whitelist_for_serdes
@public
class MultiPartitionMapping(
    BaseMultiPartitionMapping,
    PartitionMapping,
    NamedTuple(
        "_MultiPartitionMapping",
        [("downstream_mappings_by_upstream_dimension", Mapping[str, DimensionPartitionMapping])],
    ),
):
    """Defines a correspondence between two MultiPartitionsDefinitions.

    Accepts a mapping of upstream dimension name to downstream DimensionPartitionMapping, representing
    the explicit correspondence between the upstream and downstream MultiPartitions dimensions
    and the partition mapping used to calculate the downstream partitions.

    Examples:
        .. code-block:: python

            weekly_abc = MultiPartitionsDefinition(
                {
                    "abc": StaticPartitionsDefinition(["a", "b", "c"]),
                    "weekly": WeeklyPartitionsDefinition("2023-01-01"),
                }
            )
            daily_123 = MultiPartitionsDefinition(
                {
                    "123": StaticPartitionsDefinition(["1", "2", "3"]),
                    "daily": DailyPartitionsDefinition("2023-01-01"),
                }
            )

            MultiPartitionMapping(
                {
                    "abc": DimensionPartitionMapping(
                        dimension_name="123",
                        partition_mapping=StaticPartitionMapping({"a": "1", "b": "2", "c": "3"}),
                    ),
                    "weekly": DimensionPartitionMapping(
                        dimension_name="daily",
                        partition_mapping=TimeWindowPartitionMapping(),
                    )
                }
            )

    For upstream or downstream dimensions not explicitly defined in the mapping, Dagster will
    assume an `AllPartitionsMapping`, meaning that all upstream partitions in those dimensions
    will be mapped to all downstream partitions in those dimensions.

    Examples:
        .. code-block:: python

            weekly_abc = MultiPartitionsDefinition(
                {
                    "abc": StaticPartitionsDefinition(["a", "b", "c"]),
                    "daily": DailyPartitionsDefinition("2023-01-01"),
                }
            )
            daily_123 = MultiPartitionsDefinition(
                {
                    "123": StaticPartitionsDefinition(["1", "2", "3"]),
                    "daily": DailyPartitionsDefinition("2023-01-01"),
                }
            )

            MultiPartitionMapping(
                {
                    "daily": DimensionPartitionMapping(
                        dimension_name="daily",
                        partition_mapping=IdentityPartitionMapping(),
                    )
                }
            )

            # Will map `daily_123` partition key {"123": "1", "daily": "2023-01-01"} to the upstream:
            # {"abc": "a", "daily": "2023-01-01"}
            # {"abc": "b", "daily": "2023-01-01"}
            # {"abc": "c", "daily": "2023-01-01"}

    Args:
        downstream_mappings_by_upstream_dimension (Mapping[str, DimensionPartitionMapping]): A
            mapping that defines an explicit correspondence between one dimension of the upstream
            MultiPartitionsDefinition and one dimension of the downstream MultiPartitionsDefinition.
            Maps a string representing upstream dimension name to downstream DimensionPartitionMapping,
            containing the downstream dimension name and partition mapping.
    """

    def __new__(
        cls, downstream_mappings_by_upstream_dimension: Mapping[str, DimensionPartitionMapping]
    ):
        return super().__new__(
            cls,
            downstream_mappings_by_upstream_dimension=check.mapping_param(
                downstream_mappings_by_upstream_dimension,
                "downstream_mappings_by_upstream_dimension",
                key_type=str,
                value_type=DimensionPartitionMapping,
            ),
        )

    @property
    def description(self) -> str:
        return "\n ".join(
            [
                (
                    f"Upstream dimension '{upstream_dim}' mapped to downstream dimension "
                    f"'{downstream_mapping.dimension_name}' using {type(downstream_mapping.partition_mapping).__name__}."
                )
                for upstream_dim, downstream_mapping in self.downstream_mappings_by_upstream_dimension.items()
            ]
        )

    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ):
        self._check_all_dimensions_accounted_for(
            upstream_partitions_def,
            downstream_partitions_def,
        )

    def get_dimension_dependencies(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
    ) -> Sequence[DimensionDependency]:
        self._check_all_dimensions_accounted_for(
            upstream_partitions_def,
            downstream_partitions_def,
        )

        return [
            DimensionDependency(
                mapping.partition_mapping,
                upstream_dimension_name=upstream_dimension,
                downstream_dimension_name=mapping.dimension_name,
            )
            for upstream_dimension, mapping in self.downstream_mappings_by_upstream_dimension.items()
        ]

    def _check_all_dimensions_accounted_for(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ) -> None:
        if any(
            not isinstance(partitions_def, MultiPartitionsDefinition)
            for partitions_def in (upstream_partitions_def, downstream_partitions_def)
        ):
            check.failed(
                "Both partitions defs provided to a MultiPartitionMapping must be multi-partitioned"
            )

        upstream_dimension_names = {
            dim.name
            for dim in cast("MultiPartitionsDefinition", upstream_partitions_def).partitions_defs
        }
        dimension_names = {
            dim.name
            for dim in cast("MultiPartitionsDefinition", downstream_partitions_def).partitions_defs
        }

        for (
            upstream_dimension_name,
            dimension_mapping,
        ) in self.downstream_mappings_by_upstream_dimension.items():
            if upstream_dimension_name not in upstream_dimension_names:
                check.failed(
                    "Dimension mapping has an upstream dimension name that is not in the upstream "
                    "partitions def"
                )
            if dimension_mapping.dimension_name not in dimension_names:
                check.failed(
                    "Dimension mapping has a downstream dimension name that is not in the"
                    " downstream partitions def"
                )

            upstream_dimension_names.remove(upstream_dimension_name)
            dimension_names.remove(dimension_mapping.dimension_name)



from abc import ABC, abstractmethod, abstractproperty
from collections.abc import Sequence
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Optional

from dagster._annotations import public
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._record import record

if TYPE_CHECKING:
    from dagster._core.instance import DynamicPartitionsStore


@record
class UpstreamPartitionsResult:
    """Represents the result of mapping a PartitionsSubset to the corresponding
    partitions in another PartitionsDefinition.

    partitions_subset (PartitionsSubset): The resulting partitions subset that was
        mapped to. Only contains partitions for existent partitions, filtering out nonexistent partitions.
    required_but_nonexistent_subset (PartitionsSubset): A set of invalid partition keys in to_partitions_def
        that partitions in from_partitions_subset were mapped to.
    """

    partitions_subset: PartitionsSubset
    required_but_nonexistent_subset: PartitionsSubset

    @cached_property
    def required_but_nonexistent_partition_keys(self) -> Sequence[str]:
        return list(self.required_but_nonexistent_subset.get_partition_keys())


@public
class PartitionMapping(ABC):
    """Defines a correspondence between the partitions in an asset and the partitions in an asset
    that it depends on.

    Overriding PartitionMapping outside of Dagster is not supported. The abstract methods of this
    class may change at any time.
    """

    @public
    @abstractmethod
    def get_downstream_partitions_for_partitions(
        self,
        upstream_partitions_subset: PartitionsSubset,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> PartitionsSubset:
        """Returns the subset of partition keys in the downstream asset that use the data in the given
        partition key subset of the upstream asset.

        Args:
            upstream_partitions_subset (Union[PartitionKeyRange, PartitionsSubset]): The
                subset of partition keys in the upstream asset.
            downstream_partitions_def (PartitionsDefinition): The partitions definition for the
                downstream asset.
        """

    @abstractmethod
    def validate_partition_mapping(
        self,
        upstream_partitions_def: PartitionsDefinition,
        downstream_partitions_def: PartitionsDefinition | None,
    ) -> None:
        """Raises an exception if the mapping is not valid for the two partitions definitions
        due to some incompatibility between the definitions (ignoring specific keys or subsets).
        For example, a StaticPartitionMapping is invalid if both mapped partitions definitions
        are not StaticPartitionsDefinitions.
        """

    @public
    @abstractmethod
    def get_upstream_mapped_partitions_result_for_partitions(
        self,
        downstream_partitions_subset: PartitionsSubset | None,
        downstream_partitions_def: PartitionsDefinition | None,
        upstream_partitions_def: PartitionsDefinition,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> UpstreamPartitionsResult:
        """Returns a UpstreamPartitionsResult object containing the partition keys the downstream
        partitions subset was mapped to in the upstream partitions definition.

        Valid upstream partitions will be included in UpstreamPartitionsResult.partitions_subset.
        Invalid upstream partitions will be included in UpstreamPartitionsResult.required_but_nonexistent_subset.

        For example, if an upstream asset is time-partitioned and starts in June 2023, and the
        downstream asset is time-partitioned and starts in May 2023, this function would return a
        UpstreamPartitionsResult(PartitionsSubset("2023-06-01"), required_but_nonexistent_subset=PartitionsSubset("2023-05-01"))
        when downstream_partitions_subset contains 2023-05-01 and 2023-06-01.
        """

    @abstractproperty
    def description(self) -> str:
        """A human-readable description of the partition mapping, displayed in the Dagster UI."""
        raise NotImplementedError()



from typing import NamedTuple

from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes
class SerializedPartitionsSubset(NamedTuple):
    serialized_subset: str
    serialized_partitions_def_unique_id: str
    serialized_partitions_def_class_name: str

    @classmethod
    def from_subset(cls, subset: PartitionsSubset, partitions_def: PartitionsDefinition):
        return cls(
            serialized_subset=subset.serialize(),
            serialized_partitions_def_unique_id=partitions_def.get_serializable_unique_identifier(),
            serialized_partitions_def_class_name=partitions_def.__class__.__name__,
        )

    def can_deserialize(self, partitions_def: PartitionsDefinition | None) -> bool:
        if not partitions_def:
            # Asset had a partitions definition at storage time, but no longer does
            return False

        return partitions_def.can_deserialize_subset(
            self.serialized_subset,
            serialized_partitions_def_unique_id=self.serialized_partitions_def_unique_id,
            serialized_partitions_def_class_name=self.serialized_partitions_def_class_name,
        )

    def deserialize(self, partitions_def: PartitionsDefinition) -> PartitionsSubset:
        return partitions_def.deserialize_subset(self.serialized_subset)



from abc import abstractmethod
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from dagster._core.instance.instance import DagsterInstance
    from dagster._core.run_coordinator.base import RunCoordinator
    from dagster._core.scheduler.scheduler import Scheduler
    from dagster._core.storage.runs import RunStorage
    from dagster._daemon.types import DaemonHeartbeat, DaemonStatus


class DaemonMethods:
    """Mixin class containing daemon-related functionality for DagsterInstance.

    This class provides methods for daemon management, heartbeats, and status monitoring.
    All methods are implemented as instance methods that DagsterInstance inherits.
    """

    # Abstract properties that DagsterInstance provides
    @property
    @abstractmethod
    def run_storage(self) -> "RunStorage": ...

    @property
    @abstractmethod
    def is_ephemeral(self) -> bool: ...

    @property
    @abstractmethod
    def scheduler(self) -> Optional["Scheduler"]: ...

    @property
    @abstractmethod
    def run_coordinator(self) -> "RunCoordinator": ...

    @property
    @abstractmethod
    def run_monitoring_enabled(self) -> bool: ...

    @property
    @abstractmethod
    def run_retries_enabled(self) -> bool: ...

    @property
    @abstractmethod
    def auto_materialize_enabled(self) -> bool: ...

    @property
    @abstractmethod
    def auto_materialize_use_sensors(self) -> bool: ...

    @property
    @abstractmethod
    def freshness_enabled(self) -> bool: ...

    def add_daemon_heartbeat(self, daemon_heartbeat: "DaemonHeartbeat") -> None:
        """Called on a regular interval by the daemon."""
        self.run_storage.add_daemon_heartbeat(daemon_heartbeat)

    def get_daemon_heartbeats(self) -> Mapping[str, "DaemonHeartbeat"]:
        """Latest heartbeats of all daemon types."""
        return self.run_storage.get_daemon_heartbeats()

    def wipe_daemon_heartbeats(self) -> None:
        self.run_storage.wipe_daemon_heartbeats()

    def get_required_daemon_types(self) -> Sequence[str]:
        from dagster._core.run_coordinator import QueuedRunCoordinator
        from dagster._core.scheduler import DagsterDaemonScheduler
        from dagster._daemon.asset_daemon import AssetDaemon
        from dagster._daemon.auto_run_reexecution.event_log_consumer import EventLogConsumerDaemon
        from dagster._daemon.daemon import (
            BackfillDaemon,
            MonitoringDaemon,
            SchedulerDaemon,
            SensorDaemon,
        )
        from dagster._daemon.freshness import FreshnessDaemon
        from dagster._daemon.run_coordinator.queued_run_coordinator_daemon import (
            QueuedRunCoordinatorDaemon,
        )

        if self.is_ephemeral:
            return []

        daemons = [SensorDaemon.daemon_type(), BackfillDaemon.daemon_type()]
        if isinstance(self.scheduler, DagsterDaemonScheduler):
            daemons.append(SchedulerDaemon.daemon_type())
        if isinstance(self.run_coordinator, QueuedRunCoordinator):
            daemons.append(QueuedRunCoordinatorDaemon.daemon_type())
        if self.run_monitoring_enabled:
            daemons.append(MonitoringDaemon.daemon_type())
        if self.run_retries_enabled:
            daemons.append(EventLogConsumerDaemon.daemon_type())
        if self.auto_materialize_enabled or self.auto_materialize_use_sensors:
            daemons.append(AssetDaemon.daemon_type())
        if self.freshness_enabled:
            daemons.append(FreshnessDaemon.daemon_type())
        return daemons

    def get_daemon_statuses(
        self, daemon_types: Sequence[str] | None = None
    ) -> Mapping[str, "DaemonStatus"]:
        """Get the current status of the daemons. If daemon_types aren't provided, defaults to all
        required types. Returns a dict of daemon type to status.
        """
        from typing import cast

        import dagster._check as check
        from dagster._daemon.controller import get_daemon_statuses

        check.opt_sequence_param(daemon_types, "daemon_types", of_type=str)
        # Cast is safe since this mixin is only used by DagsterInstance
        return get_daemon_statuses(
            cast("DagsterInstance", self),
            daemon_types=daemon_types or self.get_required_daemon_types(),
            ignore_errors=True,
        )

    @property
    def daemon_skip_heartbeats_without_errors(self) -> bool:
        # If enabled, daemon threads won't write heartbeats unless they encounter an error. This is
        # enabled in cloud, where we don't need to use heartbeats to check if daemons are running, but
        # do need to surface errors to users. This is an optimization to reduce DB writes.
        return False



import os

from dagster_shared.utils import environ as environ


def using_dagster_dev() -> bool:
    return bool(os.getenv("DAGSTER_IS_DEV_CLI"))


def use_verbose() -> bool:
    return bool(os.getenv("DAGSTER_verbose", "1"))



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from typing import Generic

from typing_extensions import TypeVar

from dagster._annotations import public
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange

T_str = TypeVar("T_str", bound=str, default=str, covariant=True)


class PartitionsSubset(ABC, Generic[T_str]):
    """Represents a subset of the partitions within a PartitionsDefinition."""

    @property
    def is_empty(self) -> bool:
        return len(self) == 0

    @abstractmethod
    def get_partition_keys_not_in_subset(
        self, partitions_def: PartitionsDefinition[T_str]
    ) -> Iterable[T_str]: ...

    @abstractmethod
    @public
    def get_partition_keys(self) -> Iterable[T_str]: ...

    @abstractmethod
    def get_partition_key_ranges(
        self, partitions_def: PartitionsDefinition
    ) -> Sequence[PartitionKeyRange]: ...

    @abstractmethod
    def with_partition_keys(self, partition_keys: Iterable[str]) -> "PartitionsSubset[T_str]": ...

    def with_partition_key_range(
        self, partitions_def: PartitionsDefinition[T_str], partition_key_range: PartitionKeyRange
    ) -> "PartitionsSubset[T_str]":
        return self.with_partition_keys(
            partitions_def.get_partition_keys_in_range(partition_key_range)
        )

    def __or__(self, other: "PartitionsSubset") -> "PartitionsSubset":
        from dagster._core.definitions.partitions.subset.all import AllPartitionsSubset

        if self is other or other.is_empty:
            return self
        # Anything | AllPartitionsSubset = AllPartitionsSubset
        # (this assumes the two subsets are using the same partitions definition)
        if isinstance(other, AllPartitionsSubset):
            return other
        return self.with_partition_keys(other.get_partition_keys())

    def __sub__(self, other: "PartitionsSubset") -> "PartitionsSubset":
        from dagster._core.definitions.partitions.subset.all import AllPartitionsSubset

        if self is other:
            return self.empty_subset()
        if other.is_empty:
            return self
        # Anything - AllPartitionsSubset = Empty
        # (this assumes the two subsets are using the same partitions definition)
        if isinstance(other, AllPartitionsSubset):
            return self.empty_subset()
        return self.empty_subset().with_partition_keys(
            set(self.get_partition_keys()).difference(set(other.get_partition_keys()))
        )

    def __and__(self, other: "PartitionsSubset") -> "PartitionsSubset":
        from dagster._core.definitions.partitions.subset.all import AllPartitionsSubset

        if self is other:
            return self
        if other.is_empty:
            return other
        # Anything & AllPartitionsSubset = Anything
        # (this assumes the two subsets are using the same partitions definition)
        if isinstance(other, AllPartitionsSubset):
            return self
        return self.empty_subset().with_partition_keys(
            set(self.get_partition_keys()) & set(other.get_partition_keys())
        )

    @abstractmethod
    def serialize(self) -> str: ...

    @classmethod
    @abstractmethod
    def from_serialized(
        cls, partitions_def: PartitionsDefinition[T_str], serialized: str
    ) -> "PartitionsSubset[T_str]": ...

    @classmethod
    @abstractmethod
    def can_deserialize(
        cls,
        partitions_def: PartitionsDefinition,
        serialized: str,
        serialized_partitions_def_unique_id: str | None,
        serialized_partitions_def_class_name: str | None,
    ) -> bool: ...

    @abstractmethod
    def __len__(self) -> int: ...

    @abstractmethod
    def __contains__(self, value) -> bool: ...

    def empty_subset(self) -> "PartitionsSubset[T_str]": ...

    @classmethod
    @abstractmethod
    def create_empty_subset(
        cls, partitions_def: PartitionsDefinition | None = None
    ) -> "PartitionsSubset[T_str]": ...

    def to_serializable_subset(self) -> "PartitionsSubset":
        return self



import hashlib
import itertools
from collections.abc import Mapping, Sequence
from datetime import datetime
from functools import lru_cache, reduce
from typing import TYPE_CHECKING, Optional, cast

from dagster_shared.check.functions import CheckError

import dagster._check as check
from dagster._annotations import public
from dagster._core.definitions.partitions.context import (
    PartitionLoadingContext,
    partition_loading_context,
)
from dagster._core.definitions.partitions.definition.dynamic import DynamicPartitionsDefinition
from dagster._core.definitions.partitions.definition.partitions_definition import (
    PartitionsDefinition,
)
from dagster._core.definitions.partitions.definition.static import StaticPartitionsDefinition
from dagster._core.definitions.partitions.definition.time_window import (
    TimeWindowPartitionsDefinition,
)
from dagster._core.definitions.partitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.partitions.subset.default import DefaultPartitionsSubset
from dagster._core.definitions.partitions.utils.multi import (
    INVALID_STATIC_PARTITIONS_KEY_CHARACTERS,
    MULTIPARTITION_KEY_DELIMITER,
    MultiDimensionalPartitionKeyIterator,
    MultiPartitionCursor,
    MultiPartitionKey,
    PartitionDimensionDefinition,
    get_tags_from_multi_partition_key,
)
from dagster._core.definitions.partitions.utils.time_window import TimeWindow
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvalidInvocationError,
    DagsterUnknownPartitionError,
)
from dagster._core.types.pagination import PaginatedResults

if TYPE_CHECKING:
    from dagster._core.definitions.partitions.subset.partitions_subset import PartitionsSubset
    from dagster._core.instance import DynamicPartitionsStore

ALLOWED_PARTITION_DIMENSION_TYPES = (
    StaticPartitionsDefinition,
    TimeWindowPartitionsDefinition,
    DynamicPartitionsDefinition,
)


def _check_valid_partitions_dimensions(
    partitions_dimensions: Mapping[str, PartitionsDefinition],
) -> None:
    for dim_name, partitions_def in partitions_dimensions.items():
        if not any(isinstance(partitions_def, t) for t in ALLOWED_PARTITION_DIMENSION_TYPES):
            raise DagsterInvalidDefinitionError(
                f"Invalid partitions definition type {type(partitions_def)}. "
                "Only the following partitions definition types are supported: "
                f"{ALLOWED_PARTITION_DIMENSION_TYPES}."
            )
        if isinstance(partitions_def, DynamicPartitionsDefinition) and partitions_def.name is None:
            raise DagsterInvalidDefinitionError(
                "DynamicPartitionsDefinition must have a name to be used in a"
                " MultiPartitionsDefinition."
            )

        if isinstance(partitions_def, StaticPartitionsDefinition):
            if any(
                [
                    INVALID_STATIC_PARTITIONS_KEY_CHARACTERS & set(key)
                    for key in partitions_def.get_partition_keys()
                ]
            ):
                raise DagsterInvalidDefinitionError(
                    f"Invalid character in partition key for dimension {dim_name}. "
                    "A multi-partitions definition cannot contain partition keys with "
                    "the following characters: |, [, ], ,"
                )


@public
class MultiPartitionsDefinition(PartitionsDefinition[MultiPartitionKey]):
    """Takes the cross-product of partitions from two partitions definitions.

    For example, with a static partitions definition where the partitions are ["a", "b", "c"]
    and a daily partitions definition, this partitions definition will have the following
    partitions:

    2020-01-01|a
    2020-01-01|b
    2020-01-01|c
    2020-01-02|a
    2020-01-02|b
    ...

    We recommended limiting partition counts for each asset to 100,000 partitions or fewer.

    Args:
        partitions_defs (Mapping[str, PartitionsDefinition]):
            A mapping of dimension name to partitions definition. The total set of partitions will
            be the cross-product of the partitions from each PartitionsDefinition.

    Args:
        partitions_defs (Sequence[PartitionDimensionDefinition]):
            A sequence of PartitionDimensionDefinition objects, each of which contains a dimension
            name and a PartitionsDefinition. The total set of partitions will be the cross-product
            of the partitions from each PartitionsDefinition. This sequence is ordered by
            dimension name, to ensure consistent ordering of the partitions.
    """

    def __init__(self, partitions_defs: Mapping[str, PartitionsDefinition]):
        if not len(partitions_defs.keys()) == 2:
            raise DagsterInvalidInvocationError(
                "Dagster currently only supports multi-partitions definitions with 2 partitions"
                " definitions. Your multi-partitions definition has"
                f" {len(partitions_defs.keys())} partitions definitions."
            )
        check.mapping_param(
            partitions_defs, "partitions_defs", key_type=str, value_type=PartitionsDefinition
        )

        _check_valid_partitions_dimensions(partitions_defs)

        self._partitions_defs: list[PartitionDimensionDefinition] = sorted(
            [
                PartitionDimensionDefinition(name, partitions_def)
                for name, partitions_def in partitions_defs.items()
            ],
            key=lambda x: x.name,
        )

    @property
    def partitions_subset_class(self) -> type["PartitionsSubset"]:
        return DefaultPartitionsSubset

    def get_partition_keys_in_range(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        partition_key_range: PartitionKeyRange,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> Sequence[str]:
        with partition_loading_context(dynamic_partitions_store=dynamic_partitions_store):
            start: MultiPartitionKey = self.get_partition_key_from_str(partition_key_range.start)
            end: MultiPartitionKey = self.get_partition_key_from_str(partition_key_range.end)

            partition_key_sequences = [
                partition_dim.partitions_def.get_partition_keys_in_range(
                    PartitionKeyRange(
                        start.keys_by_dimension[partition_dim.name],
                        end.keys_by_dimension[partition_dim.name],
                    ),
                )
                for partition_dim in self._partitions_defs
            ]

            return [
                MultiPartitionKey(
                    {
                        self._partitions_defs[i].name: key
                        for i, key in enumerate(partition_key_tuple)
                    }
                )
                for partition_key_tuple in itertools.product(*partition_key_sequences)
            ]

    def get_serializable_unique_identifier(self) -> str:
        return hashlib.sha1(
            str(
                {
                    dim_def.name: dim_def.partitions_def.get_serializable_unique_identifier()
                    for dim_def in self.partitions_defs
                }
            ).encode("utf-8")
        ).hexdigest()

    @property
    def partition_dimension_names(self) -> list[str]:
        return [dim_def.name for dim_def in self._partitions_defs]

    @property
    def partitions_defs(self) -> Sequence[PartitionDimensionDefinition]:
        return self._partitions_defs

    def get_partitions_def_for_dimension(self, dimension_name: str) -> PartitionsDefinition:
        for dim_def in self._partitions_defs:
            if dim_def.name == dimension_name:
                return dim_def.partitions_def
        check.failed(f"Invalid dimension name {dimension_name}")

    # We override the default implementation of `has_partition_key` for performance.
    def has_partition_key(self, partition_key: MultiPartitionKey | str) -> bool:
        if isinstance(partition_key, str):
            try:
                partition_key = self.get_partition_key_from_str(partition_key)
            except CheckError:
                return False

        if partition_key.keys_by_dimension.keys() != set(self.partition_dimension_names):
            raise DagsterUnknownPartitionError(
                f"Invalid partition key {partition_key}. The dimensions of the partition key are"
                " not the dimensions of the partitions definition."
            )

        for dimension in self.partitions_defs:
            if not dimension.partitions_def.has_partition_key(
                partition_key.keys_by_dimension[dimension.name]
            ):
                return False
        return True

    # store results for repeated calls with the same current_time
    @lru_cache(maxsize=1)
    def _get_partition_keys(self, _current_time: datetime) -> Sequence[MultiPartitionKey]:
        partition_key_sequences = [
            partition_dim.partitions_def.get_partition_keys()
            for partition_dim in self._partitions_defs
        ]

        keys = [
            MultiPartitionKey(
                {self._partitions_defs[i].name: key for i, key in enumerate(partition_key_tuple)}
            )
            for partition_key_tuple in itertools.product(*partition_key_sequences)
        ]
        # in some cases, an underlying partitions definition may have keys in a format
        # that produce invalid multi-partition keys (e.g. they have a | character).
        # in these cases, we filter out the invalid keys.
        return [key for key in keys if self._is_valid_key_format(key)]

    @public
    def get_partition_keys(
        self,
        current_time: datetime | None = None,
        dynamic_partitions_store: Optional["DynamicPartitionsStore"] = None,
    ) -> Sequence[MultiPartitionKey]:
        """Returns a list of MultiPartitionKeys representing the partition keys of the
        PartitionsDefinition.

        Args:
            current_time (Optional[datetime]): A datetime object representing the current time, only
                applicable to time-based partition dimensions.
            dynamic_partitions_store (Optional[DynamicPartitionsStore]): The DynamicPartitionsStore
                object that is responsible for fetching dynamic partitions. Required when a
                dimension is a DynamicPartitionsDefinition with a name defined. Users can pass the
                DagsterInstance fetched via `context.instance` to this argument.

        Returns:
            Sequence[MultiPartitionKey]
        """
        with partition_loading_context(current_time, dynamic_partitions_store) as ctx:
            return self._get_partition_keys(_current_time=ctx.effective_dt)

    def get_paginated_partition_keys(
        self,
        context: PartitionLoadingContext,
        limit: int,
        ascending: bool,
        cursor: str | None = None,
    ) -> PaginatedResults[str]:
        """Returns a connection object that contains a list of partition keys and all the necessary
        information to paginate through them.

        Args:
            cursor: (Optional[str]): A cursor to track the progress paginating through the returned partition key results.
            limit: (Optional[int]): The maximum number of partition keys to return.

        Returns:
            PaginatedResults[MultiPartitionKey]
        """
        with partition_loading_context(new_ctx=context) as ctx:
            partition_keys = []
            iterator = MultiDimensionalPartitionKeyIterator(
                context=ctx,
                partition_defs=self._partitions_defs,
                cursor=MultiPartitionCursor.from_cursor(cursor),
                ascending=ascending,
            )
            next_cursor = cursor
            while iterator.has_next():
                partition_key = next(iterator)
                if not partition_key:
                    break

                partition_keys.append(partition_key)
                next_cursor = iterator.cursor().to_string()
                if len(partition_keys) >= limit:
                    break

            if not next_cursor:
                next_cursor = MultiPartitionCursor(last_seen_key=None).to_string()

            return PaginatedResults(
                results=partition_keys, cursor=next_cursor, has_more=iterator.has_next()
            )

    def _is_valid_key_format(self, partition_key: str) -> bool:
        """Checks if the given partition key is in the correct format for a multi-partition key
        of this MultiPartitionsDefinition.
        """
        return len(partition_key.split(MULTIPARTITION_KEY_DELIMITER)) == len(self.partitions_defs)

    def filter_valid_partition_keys(self, partition_keys: set[str]) -> set[MultiPartitionKey]:
        partition_keys_by_dimension = {
            dim.name: dim.partitions_def.get_partition_keys() for dim in self.partitions_defs
        }
        validated_partitions = set()
        for partition_key in partition_keys:
            if not self._is_valid_key_format(partition_key):
                continue

            partition_key_strs = partition_key.split(MULTIPARTITION_KEY_DELIMITER)
            multipartition_key = MultiPartitionKey(
                {dim.name: partition_key_strs[i] for i, dim in enumerate(self._partitions_defs)}
            )

            if all(
                key in partition_keys_by_dimension.get(dim, [])
                for dim, key in multipartition_key.keys_by_dimension.items()
            ):
                validated_partitions.add(partition_key)

        return validated_partitions

    def __eq__(self, other):
        return (
            isinstance(other, MultiPartitionsDefinition)
            and self.partitions_defs == other.partitions_defs
        )

    def __hash__(self):
        return hash(
            tuple(
                [
                    (partitions_def.name, partitions_def.__repr__())
                    for partitions_def in self.partitions_defs
                ]
            )
        )

    def __str__(self) -> str:
        dimension_1 = self._partitions_defs[0]
        dimension_2 = self._partitions_defs[1]
        partition_str = (
            "Multi-partitioned, with dimensions: \n"
            f"{dimension_1.name.capitalize()}: {dimension_1.partitions_def} \n"
            f"{dimension_2.name.capitalize()}: {dimension_2.partitions_def}"
        )
        return partition_str

    def __repr__(self) -> str:
        return f"{type(self).__name__}(dimensions={[str(dim) for dim in self.partitions_defs]}"

    def get_partition_key_from_str(self, partition_key_str: str) -> MultiPartitionKey:
        """Given a string representation of a partition key, returns a MultiPartitionKey object."""
        check.str_param(partition_key_str, "partition_key_str")

        partition_key_strs = partition_key_str.split(MULTIPARTITION_KEY_DELIMITER)
        check.invariant(
            len(partition_key_strs) == len(self.partitions_defs),
            f"Expected {len(self.partitions_defs)} partition keys in partition key string"
            f" {partition_key_str}, but got {len(partition_key_strs)}",
        )

        return MultiPartitionKey(
            {dim.name: partition_key_strs[i] for i, dim in enumerate(self._partitions_defs)}
        )

    def _get_primary_and_secondary_dimension(
        self,
    ) -> tuple[PartitionDimensionDefinition, PartitionDimensionDefinition]:
        # Multipartitions subsets are serialized by primary dimension. If changing
        # the selection of primary/secondary dimension, will need to also update the
        # serialization of MultiPartitionsSubsets

        time_dimensions = self._get_time_window_dims()
        if len(time_dimensions) == 1:
            primary_dimension, secondary_dimension = (
                time_dimensions[0],
                next(iter([dim for dim in self.partitions_defs if dim != time_dimensions[0]])),
            )
        else:
            primary_dimension, secondary_dimension = (
                self.partitions_defs[0],
                self.partitions_defs[1],
            )

        return primary_dimension, secondary_dimension

    @property
    def primary_dimension(self) -> PartitionDimensionDefinition:
        return self._get_primary_and_secondary_dimension()[0]

    @property
    def secondary_dimension(self) -> PartitionDimensionDefinition:
        return self._get_primary_and_secondary_dimension()[1]

    def get_tags_for_partition_key(self, partition_key: str) -> Mapping[str, str]:
        partition_key = cast("MultiPartitionKey", self.get_partition_key_from_str(partition_key))
        tags = {**super().get_tags_for_partition_key(partition_key)}
        tags.update(get_tags_from_multi_partition_key(partition_key))
        return tags

    @property
    def time_window_dimension(self) -> PartitionDimensionDefinition:
        time_window_dims = self._get_time_window_dims()
        check.invariant(
            len(time_window_dims) == 1, "Expected exactly one time window partitioned dimension"
        )
        return next(iter(time_window_dims))

    def _get_time_window_dims(self) -> list[PartitionDimensionDefinition]:
        return [
            dim
            for dim in self.partitions_defs
            if isinstance(dim.partitions_def, TimeWindowPartitionsDefinition)
        ]

    @property
    def has_time_window_dimension(self) -> bool:
        return bool(self._get_time_window_dims())

    @property
    def time_window_partitions_def(self) -> TimeWindowPartitionsDefinition:
        check.invariant(self.has_time_window_dimension, "Must have time window dimension")
        return cast(
            "TimeWindowPartitionsDefinition",
            check.inst(self.primary_dimension.partitions_def, TimeWindowPartitionsDefinition),
        )

    def time_window_for_partition_key(self, partition_key: str) -> TimeWindow:
        if not isinstance(partition_key, MultiPartitionKey):
            partition_key = self.get_partition_key_from_str(partition_key)

        time_window_dimension = self.time_window_dimension
        return cast(
            "TimeWindowPartitionsDefinition", time_window_dimension.partitions_def
        ).time_window_for_partition_key(
            cast("MultiPartitionKey", partition_key).keys_by_dimension[time_window_dimension.name]
        )

    def get_multipartition_keys_with_dimension_value(
        self, dimension_name: str, dimension_partition_key: str
    ) -> Sequence[MultiPartitionKey]:
        check.str_param(dimension_name, "dimension_name")
        check.str_param(dimension_partition_key, "dimension_partition_key")

        matching_dimensions = [
            dimension for dimension in self.partitions_defs if dimension.name == dimension_name
        ]
        other_dimensions = [
            dimension for dimension in self.partitions_defs if dimension.name != dimension_name
        ]

        check.invariant(
            len(matching_dimensions) == 1,
            f"Dimension {dimension_name} not found in MultiPartitionsDefinition with dimensions"
            f" {[dim.name for dim in self.partitions_defs]}",
        )

        partition_sequences = [
            partition_dim.partitions_def.get_partition_keys() for partition_dim in other_dimensions
        ] + [[dimension_partition_key]]

        # Names of partitions dimensions in the same order as partition_sequences
        partition_dim_names = [dim.name for dim in other_dimensions] + [dimension_name]

        return [
            MultiPartitionKey(
                {
                    partition_dim_names[i]: partition_key
                    for i, partition_key in enumerate(partitions_tuple)
                }
            )
            for partitions_tuple in itertools.product(*partition_sequences)
        ]

    def get_num_partitions(self) -> int:
        # Static partitions definitions can contain duplicate keys (will throw error in 1.3.0)
        # In the meantime, relying on get_num_partitions to handle duplicates to display
        # correct counts in the Dagster UI.
        dimension_counts = [dim.partitions_def.get_num_partitions() for dim in self.partitions_defs]
        return reduce(lambda x, y: x * y, dimension_counts, 1)
